"""Support tools useful to all fluorescence detectors.

Device specific support can be found in the module for the specific
hardware, e.g. ``dxp.py`` for XIA's XMAP, Mercury, Saturn, and
``xspress.py`` for Quantum Detectors's Xspress3 and Xspress3 mini.

"""

import logging
from collections import OrderedDict
from typing import Optional, Sequence

import numpy as np
from apstools.utils import cleanupText
from ophyd import Component as Cpt
from ophyd import Device, Kind
from ophyd.signal import DerivedSignal, InternalSignal

__all__ = ["XRFMixin"]


log = logging.getLogger(__name__)


active_kind = Kind.normal | Kind.config


class ROISum(Device):
    """Sums an ROI across multiple elements."""

    count = Cpt(InternalSignal, name="count")
    net_count = Cpt(InternalSignal, name="net_count")

    def __init__(self, suffix, *args, **kwargs):
        super().__init__(suffix, *args, **kwargs)
        device = self.parent.parent
        self.root_device = device
        self.roi_num = int(suffix.split(":")[-1].strip("roi"))
        # self.roi_num = int(roi_name.split(":")[-1])
        # Watch for changes to ROI values to compute new sums
        for roi in self._roi_signals():
            roi.count.subscribe(self._update_roi_total)
            roi.net_count.subscribe(self._update_roi_net)

    def _roi_signals(self):
        signals = []
        for mca in self.root_device.mca_records():
            roi = getattr(mca.rois, f"roi{self.roi_num}")
            signals.append(roi)
        return signals

    def _update_roi_total(self, *args, value, **kwargs):
        rois = self._roi_signals()
        total = np.sum([roi.count.get() for roi in rois])
        self.count.put(total, internal=True)

    def _update_roi_net(self, *args, value, **kwargs):
        rois = self._roi_signals()
        net = np.sum([roi.net_count.get() for roi in rois])
        self.net_count.put(net, internal=True)


def add_roi_sums(mcas, rois):
    """Add a set of signals to sum the ROI over all elements."""
    defn = OrderedDict()
    for roi in rois:
        attr = f"roi{roi}"
        defn[attr] = (
            ROISum,
            f"roi{roi}",
            {},
        )
    return defn


class MCASumMixin(Device):
    """Adds a signal that reports the sum of the *spectrum* signal."""

    total_count = Cpt(InternalSignal, kind="normal")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Update the total count when the spectrum changes
        self.spectrum.subscribe(self._update_total_count)

    def _update_total_count(self, *args, old_value, value, **kwargs):
        total = int(np.sum(value))
        self.total_count.put(total, internal=True)


class UseROISignal(DerivedSignal):
    """Check that the ``.use`` ROI signal properly mangles the label.

    It uses label mangling instead of any underlying PVs because
    different detector types don't have this feature or use it in an
    undesirable way.

    """

    sentinel_char = "~"

    def inverse(self, value):
        """Compute original signal value -> derived signal value"""
        try:
            disabled = value.startswith(self.sentinel_char)
        except AttributeError:
            disabled = True
        return not disabled

    def forward(self, value):
        """Compute derived signal value -> original signal value"""
        label = str(self._derived_from.get()).strip(self.sentinel_char)
        disable_roi = not bool(value)
        if disable_roi:
            label = f"{self.sentinel_char}{label}"
        return label


class ROIMixin(Device):
    _original_name = None
    _original_kinds = {}
    _dynamic_hint_fields = ["count"]

    def stage(self):
        self._original_name = self.name
        # Append the ROI label to the signal name
        label = str(self.label.get()).strip("~")
        label = cleanupText(label)
        old_name_base = self.name
        new_name_base = f"{self.name}_{label}"

        if label != "":
            log.debug(
                f"Mangling ROI label '{self.label.get()}' -> "
                f"'{label}' ('{new_name_base}')"
            )
            self.name = new_name_base
            # Update the device name for children
            for walk in self.walk_signals():
                walk.item.name = walk.item.name.replace(old_name_base, new_name_base)
        # Set the kind based on the user-settable ".R0_BS_HINTED" PV
        if bool(self.use.get()):
            new_kind = Kind.hinted
        else:
            new_kind = Kind.normal
        self._original_kinds = {
            fld: getattr(self, fld).kind for fld in self._dynamic_hint_fields
        }
        for fld in self._dynamic_hint_fields:
            getattr(self, fld).kind = new_kind
        return super().stage()

    def unstage(self):
        # Restore the original (pre-staged) name
        if self.name != self._original_name:
            for walk in self.walk_signals():
                if self._original_name is not None:
                    walk.item.name = walk.item.name.replace(
                        self.name, self._original_name
                    )
            self.name = self._original_name
        # Restore original signal kinds
        for fld, kind in self._original_kinds.items():
            getattr(self, fld).kind = kind
        super().unstage()


class XRFMixin(Device):
    """Properties common to all XRF detectors."""

    total_count = Cpt(InternalSignal, kind=active_kind)
    _mca_count_cache: dict

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._mca_count_cache = {}
        # Monitor the individual MCAs to keep track of their sums
        for mca in self.mca_records():
            mca.total_count.subscribe(self._update_total_count)

    def _update_total_count(self, *args, obj, value, timestamp, **kwargs):
        """Callback for summing the total counts across all MCAs."""
        # Update caches values
        self._mca_count_cache[obj] = value
        # Calculate new sum
        new_total = sum(self._mca_count_cache.values())
        self.total_count.put(new_total, internal=True, timestamp=timestamp)

    def enable_rois(
        self,
        rois: Optional[Sequence[int]] = None,
        elements: Optional[Sequence[int]] = None,
    ) -> list:
        """Include some, or all, ROIs in the list of detectors to
        read.

        elements
          A list of indices for which elements to enable. Default is
          to operate on all elements.

        rois
          A list of indices for which ROIs to enable. Default is to
          operate on all ROIs.

        Returns
        =======
        statuses
          The status object for each ROI that was changed
        """
        statuses = []

        if rois is None:
            rois = range(self.num_rois)

        if elements is None:
            elements = range(self.num_elements)

        for mca_num in elements:
            for roi_num in rois:
                roi = self.get_roi(mca_num, roi_num)
                status = roi.use.set(1)
                statuses.append(status)
        return statuses

    def disable_rois(
        self,
        rois: Optional[Sequence[int]] = None,
        elements: Optional[Sequence[int]] = None,
    ) -> list:
        """Remove some, or all, ROIs from the list of detectors to
        read.

        elements
          A list of indices for which elements to enable. Default is
          to operate on all elements.

        rois
          A list of indices for which ROIs to enable. Default is to
          operate on all ROIs.

        Returns
        =======
        statuses
          The status object for each ROI that was changed
        """
        statuses = []
        # Default to all elements and all ROIs
        if rois is None:
            rois = range(self.num_rois)

        if elements is None:
            elements = range(1, self.num_elements)
        # Go through and set the hint on requested ROIs
        for mca_num in elements:
            for roi_num in rois:
                roi = self.get_roi(mca_num, roi_num)
                status = roi.use.set(0)
                statuses.append(status)
        return statuses

    def get_roi(self, mca_num: int, roi_num: int):
        """Get a specific ROI component based on
        its MCA number and then the ROI number.
        """
        mca = getattr(self.mcas, f"mca{mca_num}")
        roi = getattr(mca.rois, f"roi{roi_num}")
        return roi

    @property
    def num_rois(self):
        n_rois = float("inf")
        for mca in self.mca_records():
            n_rois = min(n_rois, len(mca.rois.component_names))
        return n_rois

    @property
    def num_elements(self):
        return len(self.mca_records())

    def mca_records(self, mca_indices: Optional[Sequence[int]] = None):
        mcas = [
            getattr(self.mcas, m)
            for m in self.mcas.component_names
            if m.startswith("mca")
        ]
        # Filter by element index
        if mca_indices is not None:
            mcas = [
                m for m in mcas if int(m.dotted_name.split(".")[-1][3:]) in mca_indices
            ]
        return mcas


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import asyncio
import uuid
from unittest.mock import MagicMock

import pydm
import pytest
from ophyd import Signal
from ophyd_async.core import soft_signal_rw
from pydm.data_plugins import plugin_for_address
from pydm.widgets import PyDMChannel, PyDMLineEdit
from qtpy.QtCore import QObject
from qtpy.QtCore import Signal as QSignal

from firefly.pydm_plugin import HavenPlugin


def test_plugin_registered():
    plugin = plugin_for_address("haven://")
    assert isinstance(plugin, HavenPlugin)


def test_signal_connection(qapp, qtbot, sim_registry):
    # Create a signal and attach our listener
    sig = Signal(name="my_signal", value=1)
    sim_registry.register(sig)
    widget = PyDMLineEdit()
    qtbot.addWidget(widget)
    widget.channel = "haven://my_signal"
    listener = widget.channels()[0]
    # If PyDMChannel can not connect, we need to connect it ourselves
    # In PyDM > 1.5.0 this will not be neccesary as the widget will be
    # connected after we set the channel name
    if not hasattr(listener, "connect"):
        pydm.utilities.establish_widget_connections(widget)
    # Check that our widget receives the initial value
    qapp.processEvents()
    assert widget._write_access
    assert widget._connected
    assert widget.value == 1


@pytest.fixture()
async def async_signal(sim_registry):
    # Get a unique name so we don't recycle channels
    signal_name = str(uuid.uuid4())
    # Create the signal
    sig = soft_signal_rw(float, initial_value=1.0, name=signal_name)
    await sig.connect(mock=False)
    sim_registry.use_typhos = False  # Typhos doesn't work with async anyway?
    sim_registry.register(sig)
    return sig


@pytest.fixture()
async def async_channel(qapp, qtbot, async_signal):

    class SigHolder(QObject):
        send_value = QSignal(float)

    sigs = SigHolder()
    channel = PyDMChannel(
        address=f"haven://{async_signal.name}",
        value_signal=sigs.send_value,
        write_access_slot=MagicMock(),
        connection_slot=MagicMock(),
        value_slot=MagicMock(),
    )
    channel.connect()
    # Wait for metadata task to complete
    try:
        task = [
            t
            for t in asyncio.all_tasks()
            if t.get_name() == f"meta_{async_signal.name}"
        ][0]
    except IndexError:
        # Connection already established
        pass
    else:
        await task
    # Execute the test code
    try:
        yield channel
    finally:
        channel.disconnect()


@pytest.mark.asyncio
async def test_async_signal_connection(async_channel, async_signal):
    # Check that our widget receives the initial values from the signal
    channel = async_channel
    assert channel.write_access_slot.called
    assert channel.connection_slot.called
    channel.value_slot.assert_called_once_with(1)


@pytest.mark.asyncio
async def test_async_send_value(qtbot, qapp, async_channel, async_signal):
    """Can we update the ophyd signal from the channel?"""
    channel = async_channel
    signal = async_signal
    # Check our starting state
    assert (await signal.get_value()) == 1.0
    # Update the widget
    channel.value_signal.emit(2.0)
    for task in asyncio.all_tasks():
        try:
            await task
        except RuntimeError:
            pass
    # Check the updated value
    assert (await signal.get_value()) == 2.0


@pytest.mark.asyncio
async def test_async_gets_value(qtbot, qapp, async_channel, async_signal):
    """Can we update the channel from the ophyd signal?"""
    channel = async_channel
    signal = async_signal
    channel.value_slot.reset_mock()
    # Update the ophyd signal
    await signal.set(3.0)
    qapp.processEvents()
    # Check that the channel slot was updated
    channel.value_slot.assert_called_once_with(3.0)



import logging

from apstools.devices.aps_machine import ApsMachineParametersDevice
from ophyd import Component as Cpt
from ophyd import EpicsSignalRO

log = logging.getLogger(__name__)


class ApsMachine(ApsMachineParametersDevice):
    _default_read_attrs = [
        "current",
        "lifetime",
    ]
    _default_configuration_attrs = [
        "aps_cycle",
        "machine_status",
        "operating_mode",
        "shutter_status",
        "fill_number",
        "orbit_correction",
        # Removed in apstools 1.6.20
        # "global_feedback",
        # "global_feedback_h",
        # "global_feedback_v",
        "operator_messages",
    ]
    shutter_status = Cpt(EpicsSignalRO, "RF-ACIS:FePermit:Sect1To35IdM.RVAL")

    def __init__(
        self,
        prefix: str = "",
        *,
        name: str,
        labels: set[str] = {"synchrotrons"},
        **kwargs,
    ):
        super().__init__(prefix=prefix, name=name, labels=labels, **kwargs)


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



"""
Ophyd support for Stanford Research Systems 570 preamplifier from synApps

Public Structures

This device connects with the SRS570 support from synApps.
(https://github.com/epics-modules/ip/blob/master/ipApp/Db/SR570.db)

The SRS570 synApps support is part of the ``ip`` module:
https://htmlpreview.github.io/?https://raw.githubusercontent.com/epics-modules/ip/R3-6-1/documentation/swaitRecord.html

:see: https://github.com/epics-modules/ip
"""

import asyncio
import logging
import math
from collections import OrderedDict
from enum import Enum
from typing import Optional, Type

from ophyd_async.core import (
    CALCULATE_TIMEOUT,
    AsyncStatus,
    CalculatableTimeout,
    Device,
    SignalRW,
    SubsetEnum,
    T,
)
from ophyd_async.epics.signal import epics_signal_rw, epics_signal_x
from ophyd_async.epics.signal._signal import _epics_signal_backend

from .. import exceptions
from .signal import derived_signal_r, derived_signal_rw

logger = logging.getLogger(__name__)


gain_units = ["pA/V", "nA/V", "uA/V", "mA/V"]
gain_values = ["1", "2", "5", "10", "20", "50", "100", "200", "500"]
gain_modes = ["LOW NOISE", "HIGH BW"]


class Sign(str, Enum):
    PLUS = "+"
    MINUS = "-"


class Cal(str, Enum):
    CAL = "CAL"
    UNCAL = "UNCAL"


# Settling times measured from the 25-ID-C upstream I0 chamber's SR-570
# (sensitivity_value, sensitivity_unit, gain_mode): settle_time
settling_times = {
    # pA/V
    ("1", "pA/V", "HIGH BW"): 2.5,
    ("2", "pA/V", "HIGH BW"): 2,
    ("5", "pA/V", "HIGH BW"): 2.0,
    ("10", "pA/V", "HIGH BW"): 0.5,
    ("20", "pA/V", "HIGH BW"): 0.5,
    ("50", "pA/V", "HIGH BW"): 0.5,
    ("100", "pA/V", "HIGH BW"): 0.5,
    ("200", "pA/V", "HIGH BW"): 0.3,
    ("500", "pA/V", "HIGH BW"): 0.3,
    ("1", "pA/V", "LOW NOISE"): 3.0,
    ("2", "pA/V", "LOW NOISE"): 2.5,
    ("5", "pA/V", "LOW NOISE"): 2.0,
    ("10", "pA/V", "LOW NOISE"): 2.0,
    ("20", "pA/V", "LOW NOISE"): 1.75,
    ("50", "pA/V", "LOW NOISE"): 1.5,
    ("100", "pA/V", "LOW NOISE"): 1.25,
    ("200", "pA/V", "LOW NOISE"): 0.5,
    ("500", "pA/V", "LOW NOISE"): 0.5,
}
settling_times.update(
    {
        # nA/V, high bandwidth
        (gain_values[idx], "nA/V", "HIGH BW"): 0.3
        for idx in range(9)
    }
)
settling_times.update(
    {
        # nA/V, low noise
        (gain_values[idx], "nA/V", "LOW NOISE"): 0.3
        for idx in range(9)
    }
)
settling_times.update(
    {
        # μA/V, high bandwidth
        (gain_values[idx], "uA/V", "HIGH BW"): 0.3
        for idx in range(9)
    }
)
settling_times.update(
    {
        # μA/V, low noise
        (gain_values[idx], "uA/V", "LOW NOISE"): 0.3
        for idx in range(9)
    }
)
settling_times.update(
    {
        ("1", "mA/V", "HIGH BW"): 0.3,
        ("1", "mA/V", "LOW NOISE"): 0.3,
    }
)


def calculate_settle_time(gain_value: int, gain_unit: int, gain_mode: str):
    """Determine the best settle time for a given combination of parameters.

    Parameters can be strings of indexes.

    """
    # Convert indexes to string values
    try:
        gain_value = gain_values[gain_value]
    except (TypeError, IndexError):
        pass
    try:
        gain_unit = gain_units[gain_unit]
    except (TypeError, IndexError):
        pass
    try:
        gain_mode = gain_modes[gain_mode]
    except (TypeError, IndexError):
        pass
    # Get calibrated settle time, or None to use the Ophyd default
    return settling_times.get((gain_value, gain_unit, gain_mode))


class GainSignal(SignalRW):
    async def calculate_settle_time(self, value):
        signals = [
            self.parent.sensitivity_value,
            self.parent.sensitivity_unit,
            self.parent.gain_mode,
        ]
        args = [value if self is sig else (await sig.get_value()) for sig in signals]
        val, unit, mode = args
        # Resolve string values to indices if provided
        if val in gain_values:
            val = gain_values.index(val)
        if unit in gain_units:
            unit = gain_units.index(unit)
        if mode in gain_modes:
            mode = gain_modes.index(mode)
        # Low-drift mode uses the same settling times as low-noise mode
        if mode == "LOW DRIFT":
            mode = "LOW NOISE"
        # Calculate settling time
        return calculate_settle_time(gain_value=val, gain_unit=unit, gain_mode=mode)

    @AsyncStatus.wrap
    async def set(
        self, value: T, wait=True, timeout: CalculatableTimeout = CALCULATE_TIMEOUT
    ) -> AsyncStatus:
        aw = super().set(value=value, wait=wait, timeout=timeout)
        if wait:
            await aw
            settle_time = await self.calculate_settle_time(value)
            await asyncio.sleep(settle_time)


def gain_signal(
    datatype: Type[T], read_pv: str, write_pv: Optional[str] = None, name: str = ""
) -> SignalRW[T]:
    """Create a `SignalRW` for changing gain and waiting for settling time.

    The gain signals are not ready to use immediately due to the amp's
    innate RC relaxation.

    Parameters
    ----------
    datatype:
        Check that the PV is of this type
    read_pv:
        The PV to read and monitor
    write_pv:
        If given, use this PV to write to, otherwise use read_pv

    """
    backend = _epics_signal_backend(datatype, read_pv, write_pv or read_pv)
    return GainSignal(backend, name=name)


# class GainSignal(EpicsSignal):
#     """
#     A signal where the settling time depends on the pre-amp gain.

#     Used to introduce a specific settle time when setting to account
#     for the amp's RC relaxation time when changing gain.
#     """

#     def set(self, value, *, timeout=DEFAULT_WRITE_TIMEOUT, settle_time="auto"):
#         """
#         Set the value of the Signal and return a Status object.

#         If put completion is used for this EpicsSignal, the status object
#         will complete once EPICS reports the put has completed.

#         Otherwise the readback will be polled until equal to the set point
#         (as in ``Signal.set``)

#         Parameters
#         ----------

#         value : any
#             The gain value.

#         timeout : float, optional
#             Maximum time to wait.

#         settle_time: float, optional
#             Delay after ``set()`` has completed to indicate completion
#             to the caller. If ``"auto"`` (default), a reasonable settle
#             time will be chosen based on the gain mode of the pre-amp.

#         Returns
#         -------
#         st : Status

#         .. seealso::
#             * Signal.set
#             * EpicsSignal.set

#         """
#         # Determine optimal settling time.
#         if settle_time == "auto":
#             signals = [self.parent.sensitivity_value, self.parent.sensitivity_unit, self.parent.gain_mode]
#             args = [value if self is sig else sig.get() for sig in signals]
#             val, unit, mode = args
#             # Resolve string values to indices if provided
#             if val in gain_values:
#                 val = gain_values.index(val)
#             if unit in gain_units:
#                 unit = gain_units.index(unit)
#             if mode in gain_modes:
#                 mode = gain_modes.index(mode)
#             # Low-drift mode uses the same settling times as low-noise mode
#             if mode == "LOW DRIFT":
#                 mode = "LOW NOISE"
#             # Calculate settling time
#             _settle_time = calculate_settle_time(gain_value=val, gain_unit=unit, gain_mode=mode)
#         else:
#             _settle_time = settle_time
#         return super().set(value, timeout=timeout, settle_time=_settle_time)


class SRS570PreAmplifier(Device):
    """Ophyd-async support for Stanford Research Systems 570 preamp."""

    offset_difference = -3  # How many levels higher should the offset be

    class FilterType(str, Enum):
        NO_FILTER = "No filter"
        _6DB_HIGHPASS = "6 dB highpass"
        _12DB_HIGHPASS = "12 dB highpass"
        _6DB_BANDPASS = "6 dB bandpass"
        _6DB_LOWPASS = "6 dB lowpass"
        _12DB_LOWPASS = "6 dB lowpass"

    class GainMode(str, Enum):
        LOW_NOISE = "LOW NOISE"
        HIGH_BW = "HIGH BW"
        LOW_DRIFT = "LOW DRIFT"

    class SensValue(str, Enum):
        _1 = "1"
        _2 = "2"
        _5 = "5"
        _10 = "10"
        _20 = "20"
        _50 = "50"
        _100 = "100"
        _200 = "200"
        _500 = "500"

    class SensUnit(str, Enum):
        pA_V = "pA/V"
        nA_V = "nA/V"
        uA_V = "uA/V"
        mA_V = "mA/V"

    class OffsetUnit(str, Enum):
        pA = "pA"
        nA = "nA"
        uA = "uA"
        mA = "mA"

    def __init__(self, prefix: str, name: str = ""):
        """
        Update the gain when the sensitivity changes.
        """
        self.sensitivity_value = gain_signal(self.SensValue, f"{prefix}sens_num")
        self.sensitivity_unit = gain_signal(self.SensUnit, f"{prefix}sens_unit")

        self.offset_on = epics_signal_rw(bool, f"{prefix}offset_on")
        self.offset_sign = epics_signal_rw(Sign, f"{prefix}offset_sign")
        self.offset_value = epics_signal_rw(self.SensValue, f"{prefix}offset_num")
        self.offset_unit = epics_signal_rw(self.OffsetUnit, f"{prefix}offset_unit")
        self.offset_fine = epics_signal_rw(int, f"{prefix}off_u_put")
        self.offset_cal = epics_signal_rw(Cal, f"{prefix}offset_cal")

        self.set_all = epics_signal_x(f"{prefix}init.PROC")

        self.bias_value = epics_signal_rw(int, f"{prefix}bias_put")
        self.bias_on = epics_signal_rw(bool, f"{prefix}bias_on")

        self.filter_type = epics_signal_rw(
            SubsetEnum[
                "  No filter",
                " 6 dB highpass",
                "12 dB highpass",
                " 6 dB bandpass",
                " 6 dB lowpass",
                "12 dB lowpass",
            ],
            f"{prefix}filter_type",
        )
        self.filter_reset = epics_signal_x(f"{prefix}filter_reset.PROC")
        self.filter_lowpass = epics_signal_rw(
            SubsetEnum[
                "  0.03 Hz",
                "  0.1 Hz",
                "  0.3 Hz",
                "  1   Hz",
                "  3   Hz",
                " 10   Hz",
                " 30   Hz",
                "100   Hz",
                "300   Hz",
                "  1   kHz",
                "  3   kHz",
                " 10   kHz",
                " 30   kHz",
                "100   kHz",
                "300   kHz",
                "  1   MHz",
            ],
            f"{prefix}low_freq",
        )
        self.filter_highpass = epics_signal_rw(
            SubsetEnum[
                "  0.03 Hz",
                "  0.1 Hz",
                "  0.3 Hz",
                "  1   Hz",
                "  3   Hz",
                " 10   Hz",
                " 30   Hz",
                "100   Hz",
                "300   Hz",
                "  1   kHz",
                "  3   kHz",
                " 10   kHz",
            ],
            f"{prefix}high_freq",
        )

        self.gain_mode = gain_signal(self.GainMode, f"{prefix}gain_mode")
        self.invert = epics_signal_rw(bool, f"{prefix}invert_on")
        self.blank = epics_signal_rw(bool, f"{prefix}blank_on")

        # Gain signals derived from the sensitivity signals
        sens_signals = {
            "sens_value": self.sensitivity_value,
            "sens_unit": self.sensitivity_unit,
        }
        self.gain = derived_signal_r(
            float,
            derived_from=sens_signals,
            inverse=self._gain_from_sensitivity,
            units="V A⁻",
        )
        self.gain_db = derived_signal_r(
            float, derived_from={"gain": self.gain}, inverse=self._dB, units="dB"
        )
        level_signals = dict(
            offset_value=self.offset_value, offset_unit=self.offset_unit, **sens_signals
        )
        self.gain_level = derived_signal_rw(
            int,
            derived_from=level_signals,
            forward=self._from_gain_level,
            inverse=self._to_gain_level,
        )
        super().__init__(name=name)

    def _gain_from_sensitivity(self, values, *, sens_value, sens_unit):
        """
        Amplifier gain (V/A), as floating-point number.
        """
        # Convert the sensitivity to a proper number
        val = float(values[sens_value])
        # Determine multiplier based on the gain unit
        amps = {
            "pA": 1e-12,
            "nA": 1e-9,
            "uA": 1e-6,
            "mA": 1e-3,
        }
        multiplier = amps[values[sens_unit].split("/")[0]]
        inverse_gain = val * multiplier
        return 1 / inverse_gain

    def _dB(self, values, *, gain):
        """Convert a gain to be in decibels."""
        try:
            return 10 * math.log10(values[gain])
        except ValueError:
            return float("nan")

    async def _from_gain_level(
        self, value, *, sens_value, sens_unit, offset_value, offset_unit
    ):
        """Compute the sensitivity settings for a given level of gain."""
        # Determine new values
        new_level = 27 - value
        new_offset = max(new_level + self.offset_difference, 0)
        # Check for out of bounds
        lmin, lmax = (0, 27)
        msg = (
            f"Cannot set {self.name} outside range ({lmin}, {lmax}), received"
            f" {new_level}."
        )
        if new_level < lmin:
            raise exceptions.GainOverflow(msg)
        elif new_level > lmax:
            raise exceptions.GainOverflow(msg)
        # Return calculated gain and offset
        result = OrderedDict()
        result[sens_unit] = self._level_to_unit(new_level)
        result[sens_value] = self._level_to_value(new_level)
        result[offset_value] = self._level_to_value(new_offset)
        result[offset_unit] = self._level_to_unit(new_offset).split("/")[0]
        return result

    def _level_to_value(self, level):
        """Convert a gain index level to its string value."""
        values = [v.value for v in self.SensValue]
        return values[int(level) % len(values)]

    def _level_to_unit(self, level):
        """Convert a gain index level to its string unit."""
        units = [u.value for u in self.SensUnit]
        values = [v.value for v in self.SensValue]
        return units[int(level / len(values))]

    def _sensitivity_to_level(self, value: str, unit: str) -> int:
        """Convert a gain value and unit to its level index."""
        values = [v.value for v in self.SensValue]
        units = [u.value for u in self.SensUnit]
        value_idx = values.index(value)
        unit_idx = units.index(unit)
        # Determine sensitivity level
        new_level = value_idx + unit_idx * len(values)
        # Convert to gain by inverting
        new_level = 27 - new_level
        return new_level

    def _to_gain_level(
        self, values, *, sens_value, sens_unit, offset_value=None, offset_unit=None
    ):
        """Compute the level of gain for given sensitivity settings."""
        # from pprint import pprint
        # pprint(values)
        return self._sensitivity_to_level(values[sens_value], values[sens_unit])


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2024, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import logging
import re  # noqa: F401

import bluesky.preprocessors as bpp  # noqa: F401
import databroker  # noqa: F401
from bluesky.plan_stubs import (  # noqa: F401
    abs_set,
    mv,
    mvr,
    null,
    pause,
    rel_set,
    sleep,
    stop,
)
from bluesky.plans import (  # noqa: F401
    count,
    grid_scan,
    list_scan,
    rel_grid_scan,
    rel_list_scan,
    rel_scan,
    scan,
    scan_nd,
)
from bluesky.run_engine import call_in_bluesky_event_loop
from ophyd_async.core import NotConnected

# Import plans
from haven import beamline  # noqa: F401
from haven import (  # noqa: F401
    align_pitch2,
    align_slits,
    auto_gain,
    energy_scan,
    knife_scan,
    recall_motor_position,
    record_dark_current,
    set_energy,
    xafs_scan,
)
from haven.run_engine import run_engine  # noqa: F401

log = logging.getLogger(__name__)

# Create a run engine without all the bells and whistles
RE = run_engine(connect_databroker=False, use_bec=False)

# Import devices
try:
    call_in_bluesky_event_loop(beamline.load())
except NotConnected as exc:
    log.exception(exc)
for cpt in beamline.registry._objects_by_name.values():
    # Replace spaces and other illegal characters in variable name
    # name = re.sub('\W|^(?=\d)','_', cpt.name)
    # Add the device as a variable in module's globals
    globals().setdefault(cpt.name, cpt)



import asyncio

from ophyd_async.core import ConfigSignal, Device, StandardReadable, SubsetEnum
from ophyd_async.epics.signal import epics_signal_r, epics_signal_rw, epics_signal_x

from ..typing import StrEnum


class AlarmStatus(StrEnum):
    NO_ALARM = "NO_ALARM"
    READ = "READ"
    WRITE = "WRITE"
    HIHI = "HIHI"
    HIGH = "HIGH"
    LOLO = "LOLO"
    LOW = "LOW"
    STATE = "STATE"
    COS = "COS"
    COMM = "COMM"
    TIMEOUT = "TIMEOUT"
    HWLIMIT = "HWLIMIT"
    CALC = "CALC"
    SCAN = "SCAN"
    LINK = "LINK"
    SOFT = "SOFT"
    # BAD_SUB = "BAD_SUB"
    # UDF = "UDF"
    # DISABLE = "DISABLE"
    # SIMM = "SIMM"
    # READ_ACCESS = "READ_ACCESS"
    # WRITE_ACCESS = "WRITE_ACCESS"


class AlarmSeverity(StrEnum):
    NO_ALARM = "NO_ALARM"
    MINOR = "MINOR"
    MAJOR = "MAJOR"
    INVALID = "INVALID"


class EpicsRecordDeviceCommonAll(StandardReadable):
    """
    Many of the fields common to all EPICS records.

    Some fields are not included because they are not interesting to
    an EPICS client or are already provided in other support.
    """

    # The valid options are specific to the record type
    # Subclasses should set this properly
    DeviceType = SubsetEnum["None"]

    class ScanInterval(StrEnum):
        PASSIVE = "Passive"
        EVENT = "Event"
        IO_INTR = "I/O Intr"
        SCAN_10 = "10 second"
        SCAN_5 = "5 second"
        SCAN_2 = "2 second"
        SCAN_1 = "1 second"
        SCAN_0_5 = ".5 second"
        SCAN_0_2 = ".2 second"
        SCAN_0_1 = ".1 second"

    # Config signals
    def __init__(self, prefix: str, name: str = ""):
        with self.add_children_as_readables(ConfigSignal):
            self.description = epics_signal_rw(str, f"{prefix}.DESC")
            self.scanning_rate = epics_signal_rw(self.ScanInterval, f"{prefix}.SCAN")
            self.device_type = epics_signal_r(self.DeviceType, f"{prefix}.DTYP")
        # Other signals, not included in read
        self.disable_value = epics_signal_rw(int, f"{prefix}.DISV")
        self.scan_disable_input_link_value = epics_signal_rw(int, f"{prefix}.DISA")
        self.scan_disable_value_input_link = epics_signal_rw(str, f"{prefix}.SDIS")
        self.forward_link = epics_signal_rw(str, f"{prefix}.FLNK")

        self.alarm_status = epics_signal_r(AlarmStatus, f"{prefix}.STAT")
        self.alarm_severity = epics_signal_r(AlarmSeverity, f"{prefix}.SEVR")
        self.new_alarm_status = epics_signal_r(AlarmStatus, f"{prefix}.NSTA")
        self.new_alarm_severity = epics_signal_r(AlarmSeverity, f"{prefix}.NSEV")
        self.disable_alarm_severity = epics_signal_rw(AlarmSeverity, f"{prefix}.DISS")
        self.processing_active = epics_signal_r(int, f"{prefix}.PACT")
        self.process_record = epics_signal_x(f"{prefix}.PROC")
        self.trace_processing = epics_signal_rw(int, f"{prefix}.TPRO")

        super().__init__(name=name)


class EpicsSynAppsRecordEnableMixin(Device):
    """Supports ``{PV}Enable`` feature from user databases."""

    def __init__(self, prefix, name=""):
        with self.add_children_as_readables(ConfigSignal):
            self.enable = epics_signal_rw(int, "Enable")
        super().__init__(name=name)

    async def reset(self):
        """set all fields to default values"""
        await asyncio.gather(
            self.enable.set(self.enable.enum_strs[1]), super().reset()  # Enable
        )


class EpicsRecordInputFields(EpicsRecordDeviceCommonAll):
    """
    Some fields common to EPICS input records.
    """

    def __init__(self, prefix: str, name: str = ""):
        with self.add_children_as_readables(ConfigSignal):
            self.input_link = epics_signal_rw(str, f"{prefix}.INP")
        super().__init__(prefix=prefix, name=name)


class EpicsRecordOutputFields(EpicsRecordDeviceCommonAll):
    """
    Some fields common to EPICS output records.
    """

    class ModeSelect(StrEnum):
        SUPERVISORY = "supervisory"
        CLOSED_LOOP = "closed_loop"

    def __init__(self, prefix: str, name: str = ""):
        with self.add_children_as_readables(ConfigSignal):
            self.output_link = epics_signal_rw(str, f"{prefix}.OUT")
            self.desired_output_location = epics_signal_rw(str, f"{prefix}.DOL")
            self.output_mode_select = epics_signal_rw(self.ModeSelect, f"{prefix}.OMSL")
        super().__init__(prefix=prefix, name=name)



import asyncio
import math

import pytest
from ophyd_async.core import Device, get_mock_put
from ophyd_async.core._signal import soft_signal_rw
from ophyd_async.epics.signal import epics_signal_x

from haven.devices.signal import derived_signal_rw, derived_signal_x


def radius(x, y):
    return (x**2 + y**2) ** 0.5


def angle(x, y):
    return math.atan2(y, x)


# Forward transforms
async def angle_to_position(value, *, x, y):
    x_, y_ = await asyncio.gather(x.get_value(), y.get_value())
    r = radius(x_, y_)
    return {
        x: r * math.cos(value),
        y: r * math.sin(value),
    }


async def radius_to_position(r, *, x, y):
    x_, y_ = await asyncio.gather(x.get_value(), y.get_value())
    theta = angle(x_, y_)
    return {
        x: r * math.cos(theta),
        y: r * math.sin(theta),
    }


# Inverse transforms
def position_to_angle(values, *, x, y):
    return angle(values[x], values[y])


def position_to_radius(values, *, x, y):
    return radius(values[x], values[y])


class Goniometer(Device):
    def __init__(self, prefix, name=""):
        self.x = soft_signal_rw(float, initial_value=0)
        self.y = soft_signal_rw(float, initial_value=0)
        real_signals = {"x": self.x, "y": self.y}
        self.angle = derived_signal_rw(
            float,
            derived_from=real_signals,
            forward=angle_to_position,
            inverse=position_to_angle,
            initial_value=0,
        )
        self.radius = derived_signal_rw(
            float,
            derived_from=real_signals,
            forward=radius_to_position,
            inverse=position_to_radius,
            initial_value=0,
        )
        super().__init__(name=name)


@pytest.fixture()
async def device():
    dev = Goniometer("255idcVME:", name="goniometer")
    await dev.connect(mock=False)
    return dev


@pytest.mark.asyncio
async def test_derived_forward(device):
    await device.radius.set(2)
    await device.angle.set(math.pi / 4)
    assert await device.x.get_value() == pytest.approx(2 / math.sqrt(2))
    assert await device.y.get_value() == pytest.approx(2 / math.sqrt(2))


@pytest.mark.asyncio
async def test_derived_defaults(device):
    """Does the derived signal report the derived value by default."""
    # Set up a pair of signals
    real_sig0 = soft_signal_rw(float)
    real_sig1 = soft_signal_rw(float)
    derived_sig = derived_signal_rw(
        float, derived_from={"real0": real_sig0, "real1": real_sig1}
    )
    await asyncio.gather(
        real_sig0.connect(), real_sig1.connect(), derived_sig.connect()
    )
    # Check that the signals are linked going forward...
    await derived_sig.set(3.3)
    assert (await real_sig0.get_value()) == 3.3
    assert (await real_sig1.get_value()) == 3.3
    # ...and backward
    await real_sig0.set(5.1)
    await real_sig1.set(5.3)
    assert (await derived_sig.get_value()) == pytest.approx(5.2)


@pytest.mark.asyncio
async def test_derived_inverse(device):
    await device.x.set(2 / math.sqrt(2))
    await device.y.set(2 / math.sqrt(2))
    assert await device.angle.get_value() == pytest.approx(math.pi / 4)
    assert await device.radius.get_value() == pytest.approx(2)


@pytest.mark.asyncio
async def test_derived_reading(device):
    await device.x.set(2 / math.sqrt(2))
    await device.y.set(2 / math.sqrt(2))
    reading = await device.radius.read()
    assert reading[device.radius.name]["value"] == pytest.approx(2)


@pytest.mark.asyncio
async def test_derived_subscribe(device):
    last_reading = {}
    listener_is_done = asyncio.Event()

    def listener(new_reading):
        last_reading.update(new_reading)
        listener_is_done.set()

    await device.y.set(1 / math.sqrt(2))
    device.angle.subscribe(listener)
    listener_is_done = asyncio.Event()
    # Now update the device
    await device.x.set(1 / math.sqrt(2))
    await asyncio.wait_for(listener_is_done.wait(), timeout=1)
    # Check that the listener fired
    assert last_reading != {}
    assert last_reading[device.angle.name]["value"] == pytest.approx(math.pi / 4)


@pytest.mark.asyncio
async def test_derived_source(device):
    desc = await device.angle.describe()
    assert desc[device.angle.name]["source"] == "soft://goniometer-angle(x,y)"


@pytest.mark.asyncio
async def test_signal_x_trigger(device):
    signal = epics_signal_x("255idcVME:start")
    derived = derived_signal_x(derived_from={"signal": signal})
    await signal.connect(mock=True)
    await derived.connect(mock=True)
    # Now trigger the parent
    mocked_put = get_mock_put(signal)
    await derived.trigger()
    mocked_put.assert_called_once_with(None, wait=True, timeout=10.0)



import asyncio
import logging
from collections import Counter
from contextlib import contextmanager
from functools import wraps
from itertools import count
from typing import Mapping, Optional, Sequence

import numpy as np
import qtawesome as qta
import yaml
from matplotlib.colors import TABLEAU_COLORS
from pandas.api.types import is_numeric_dtype
from pyqtgraph import GraphicsLayoutWidget, ImageView, PlotItem, PlotWidget
from qasync import asyncSlot
from qtpy.QtCore import Qt, Signal
from qtpy.QtGui import QStandardItem, QStandardItemModel
from qtpy.QtWidgets import QFileDialog, QWidget

from firefly import display
from firefly.run_client import DatabaseWorker

log = logging.getLogger(__name__)


colors = list(TABLEAU_COLORS.values())


def cancellable(fn):
    @wraps(fn)
    async def inner(*args, **kwargs):
        try:
            return await fn(*args, **kwargs)
        except asyncio.exceptions.CancelledError:
            log.warning(f"Cancelled task {fn}")

    return inner


class ExportDialog(QFileDialog):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setFileMode(QFileDialog.FileMode.AnyFile)
        self.setAcceptMode(QFileDialog.AcceptSave)

    def ask(self, mimetypes: Optional[Sequence[str]] = None):
        """Get the name of the file to save for exporting."""
        self.setMimeTypeFilters(mimetypes)
        # Show the file dialog
        if self.exec_() == QFileDialog.Accepted:
            return self.selectedFiles()
        else:
            return None


class FiltersWidget(QWidget):
    returnPressed = Signal()

    def keyPressEvent(self, event):
        super().keyPressEvent(event)
        # Check for return keys pressed
        if event.key() in [Qt.Key_Enter, Qt.Key_Return]:
            self.returnPressed.emit()


class Browser1DPlotItem(PlotItem):
    hover_coords_changed = Signal(str)

    def hoverEvent(self, event):
        super().hoverEvent(event)
        if event.isExit():
            self.hover_coords_changed.emit("NaN")
            return
        # Get data coordinates from event
        pos = event.scenePos()
        data_pos = self.vb.mapSceneToView(pos)
        pos_str = f"({data_pos.x():.3f}, {data_pos.y():.3f})"
        self.hover_coords_changed.emit(pos_str)


class BrowserMultiPlotWidget(GraphicsLayoutWidget):
    _multiplot_items: Mapping

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._multiplot_items = {}

    def multiplot_items(self, n_cols: int = 3):
        view = self
        item0 = None
        for idx in count():
            row = int(idx / n_cols)
            col = idx % n_cols
            # Make a new plot item if one doesn't exist
            if (row, col) not in self._multiplot_items:
                self._multiplot_items[(row, col)] = view.addPlot(row=row, col=col)
            new_item = self._multiplot_items[(row, col)]
            # Link the X-axes together
            if item0 is None:
                item0 = new_item
            else:
                new_item.setXLink(item0)
            # Resize the viewing area to fit the contents
            width = view.width()
            plot_width = width / n_cols
            # view.resize(int(width), int(plot_width * row))
            view.setFixedHeight(1200)
            yield new_item

    def plot_runs(self, runs: Mapping, xsignal: str):
        """Take loaded run data and plot small multiples.

        Parameters
        ==========
        runs
          Dictionary with pandas series for each curve. The keys
          should be the curve labels, the series' indexes are the x
          values and the series' values are the y data.
        xsignal
          The name of the signal to use for the common horizontal
          axis.

        """
        # Use all the data columns as y signals
        ysignals = []
        for run in runs.values():
            ysignals.extend(run.columns)
        # Remove the x-signal from the list of y signals
        ysignals = sorted(list(dict.fromkeys(ysignals)))
        # Plot the runs
        self.clear()
        self._multiplot_items = {}
        for label, data in runs.items():
            # Figure out which signals to plot
            try:
                xdata = data[xsignal]
            except KeyError:
                log.warning(f"Cannot plot x='{xsignal}' for {list(data.keys())}")
                continue
            # Plot each y signal on a separate plot
            for ysignal, plot_item in zip(ysignals, self.multiplot_items()):
                try:
                    if is_numeric_dtype(data[ysignal]):
                        plot_item.plot(xdata, data[ysignal])
                except KeyError:
                    log.warning(f"No signal {ysignal} in data.")
                else:
                    log.debug(f"Plotted {ysignal} vs. {xsignal} for {data}")
                plot_item.setTitle(ysignal)


class Browser1DPlotWidget(PlotWidget):
    def __init__(self, parent=None, background="default", plotItem=None, **kargs):
        plot_item = Browser1DPlotItem(**kargs)
        super().__init__(parent=parent, background=background, plotItem=plot_item)

    def plot_runs(self, runs: Mapping, ylabel="", xlabel=""):
        """Take loaded run data and plot it.

        Parameters
        ==========
        runs
          Dictionary with pandas series for each curve. The keys
          should be the curve labels, the series' indexes are the x
          values and the series' values are the y data.

        """
        plot_item = self.getPlotItem()
        plot_item.clear()
        # Plot this run's data
        cursor_needed = True
        for idx, (label, series) in enumerate(runs.items()):
            color = colors[idx % len(colors)]
            plot_item.plot(
                x=series.index,
                y=series.values,
                pen=color,
                name=label,
                clear=False,
            )
            # Cursor to drag around on the data
            if cursor_needed:
                plot_item.addLine(
                    x=np.median(series.index), movable=True, label="{value:.3f}"
                )
                cursor_needed = False
        # Axis formatting
        plot_item.setLabels(left=ylabel, bottom=xlabel)


class Browser2DPlotWidget(ImageView):
    """A plot widget for 2D maps."""

    def __init__(self, *args, view=None, **kwargs):
        if view is None:
            view = PlotItem()
        super().__init__(*args, view=view, **kwargs)

    def plot_runs(
        self, runs: Mapping, xlabel: str = "", ylabel: str = "", extents=None
    ):
        """Take loaded 2D or 3D mapping data and plot it.

        Parameters
        ==========
        runs
          Dictionary with pandas series for each curve. The keys
          should be the curve labels, the series' indexes are the x
          values and the series' values are the y data.
        xlabel
          The label for the horizontal axis.
        ylabel
          The label for the vertical axis.
        extents
          Spatial extents for the map as ((-y, +y), (-x, +x)).

        """
        images = np.asarray(list(runs.values()))
        # Combine the different runs into one image
        # To-do: make this respond to the combobox selection
        image = np.mean(images, axis=0)
        # To-do: Apply transformations

        # # Plot the image
        if 2 <= image.ndim <= 3:
            self.setImage(image.T, autoRange=False)
        else:
            log.info(f"Could not plot image of dataset with shape {image.shape}.")
            return
        # Determine the axes labels
        self.view.setLabel(axis="bottom", text=xlabel)
        self.view.setLabel(axis="left", text=ylabel)
        # Set axes extent
        yextent, xextent = extents
        x = xextent[0]
        y = yextent[0]
        w = xextent[1] - xextent[0]
        h = yextent[1] - yextent[0]
        self.getImageItem().setRect(x, y, w, h)


class RunBrowserDisplay(display.FireflyDisplay):
    runs_model: QStandardItemModel
    _run_col_names: Sequence = [
        "Plan",
        "Sample",
        "Edge",
        "E0",
        "Exit Status",
        "Datetime",
        "UID",
        "Proposal",
        "ESAF",
        "ESAF Users",
    ]
    _multiplot_items = {}

    selected_runs: list
    _running_db_tasks: Mapping

    export_dialog: Optional[ExportDialog] = None

    # Counter for keeping track of UI hints for long DB hits
    _busy_hinters: Counter

    def __init__(self, root_node=None, args=None, macros=None, **kwargs):
        super().__init__(args=args, macros=macros, **kwargs)
        self.selected_runs = []
        self._running_db_tasks = {}
        self._busy_hinters = Counter()
        self.db = DatabaseWorker(catalog=root_node)
        # Load the list of all runs for the selection widget
        self.db_task(self.load_runs(), name="init_load_runs")
        # Load the list of filters' field values into the comboboxes
        self.db_task(self.update_combobox_items(), name="update_combobox_items")

    def db_task(self, coro, name="default task"):
        """Executes a co-routine as a database task. Existing database
        tasks get cancelled.

        """
        # Check for existing tasks
        has_previous_task = name in self._running_db_tasks.keys()
        task_is_running = has_previous_task and not self._running_db_tasks[name].done()
        if task_is_running:
            self._running_db_tasks[name].cancel("New database task started.")
        # Wait on this task to be done
        new_task = asyncio.ensure_future(coro)
        self._running_db_tasks[name] = new_task
        return new_task

    @asyncSlot()
    async def reload_runs(self):
        """A simple wrapper to make load_runs a slot."""
        await self.load_runs()

    @cancellable
    async def load_runs(self):
        """Get the list of available runs based on filters."""
        with self.busy_hints(run_widgets=True, run_table=True, filter_widgets=False):
            runs = await self.db_task(
                self.db.load_all_runs(self.filters()),
                name="load all runs",
            )
            # Update the table view data model
            self.runs_model.clear()
            self.runs_model.setHorizontalHeaderLabels(self._run_col_names)
            for run in runs:
                items = [QStandardItem(val) for val in run.values()]
                self.ui.runs_model.appendRow(items)
            # Adjust the layout of the data table
            sort_col = self._run_col_names.index("Datetime")
            self.ui.run_tableview.sortByColumn(sort_col, Qt.DescendingOrder)
            self.ui.run_tableview.resizeColumnsToContents()
            # Let slots know that the model data have changed
            self.runs_total_label.setText(str(self.ui.runs_model.rowCount()))

    def clear_filters(self):
        self.ui.filter_proposal_combobox.setCurrentText("")
        self.ui.filter_esaf_combobox.setCurrentText("")
        self.ui.filter_sample_combobox.setCurrentText("")
        self.ui.filter_exit_status_combobox.setCurrentText("")
        self.ui.filter_current_proposal_checkbox.setChecked(False)
        self.ui.filter_current_esaf_checkbox.setChecked(False)
        self.ui.filter_plan_combobox.setCurrentText("")
        self.ui.filter_full_text_lineedit.setText("")
        self.ui.filter_edge_combobox.setCurrentText("")
        self.ui.filter_user_combobox.setCurrentText("")

    async def update_combobox_items(self):
        """"""
        with self.busy_hints(run_table=False, run_widgets=False, filter_widgets=True):
            fields = await self.db.load_distinct_fields()
            for field_name, cb in [
                ("proposal_users", self.ui.filter_proposal_combobox),
                ("proposal_id", self.ui.filter_user_combobox),
                ("esaf_id", self.ui.filter_esaf_combobox),
                ("sample_name", self.ui.filter_sample_combobox),
                ("plan_name", self.ui.filter_plan_combobox),
                ("edge", self.ui.filter_edge_combobox),
            ]:
                if field_name in fields.keys():
                    old_text = cb.currentText()
                    cb.clear()
                    cb.addItems(fields[field_name])
                    cb.setCurrentText(old_text)

    @asyncSlot()
    @cancellable
    async def sleep_slot(self):
        await self.db_task(self.print_sleep())

    async def print_sleep(self):
        with self.busy_hints(run_widgets=True, run_table=True, filter_widgets=True):
            label = self.ui.sleep_label
            label.setText(f"3...")
            await asyncio.sleep(1)
            old_text = label.text()
            label.setText(f"{old_text}2...")
            await asyncio.sleep(1)
            old_text = label.text()
            label.setText(f"{old_text}1...")
            await asyncio.sleep(1)
            old_text = label.text()
            label.setText(f"{old_text}done!")

    def customize_ui(self):
        self.load_models()
        # Setup controls for select which run to show

        self.ui.run_tableview.selectionModel().selectionChanged.connect(
            self.update_selected_runs
        )
        self.ui.refresh_runs_button.setIcon(qta.icon("fa5s.sync"))
        self.ui.refresh_runs_button.clicked.connect(self.reload_runs)
        # Sleep controls for testing async timing
        self.ui.sleep_button.clicked.connect(self.sleep_slot)
        # Respond to changes in displaying the 1d plot
        self.ui.signal_y_combobox.currentTextChanged.connect(self.update_1d_plot)
        self.ui.signal_x_combobox.currentTextChanged.connect(self.update_1d_plot)
        self.ui.signal_r_combobox.currentTextChanged.connect(self.update_1d_plot)
        self.ui.signal_r_checkbox.stateChanged.connect(self.update_1d_plot)
        self.ui.logarithm_checkbox.stateChanged.connect(self.update_1d_plot)
        self.ui.invert_checkbox.stateChanged.connect(self.update_1d_plot)
        self.ui.gradient_checkbox.stateChanged.connect(self.update_1d_plot)
        self.ui.plot_1d_hints_checkbox.stateChanged.connect(self.update_1d_signals)
        self.ui.plot_multi_hints_checkbox.stateChanged.connect(
            self.update_multi_signals
        )
        self.ui.plot_multi_hints_checkbox.stateChanged.connect(self.update_multi_plot)
        # Respond to changes in displaying the 2d plot
        self.ui.signal_value_combobox.currentTextChanged.connect(self.update_2d_plot)
        self.ui.logarithm_checkbox_2d.stateChanged.connect(self.update_2d_plot)
        self.ui.invert_checkbox_2d.stateChanged.connect(self.update_2d_plot)
        self.ui.gradient_checkbox_2d.stateChanged.connect(self.update_2d_plot)
        self.ui.plot_2d_hints_checkbox.stateChanged.connect(self.update_2d_signals)
        # Respond to filter controls getting updated
        self.ui.filters_widget.returnPressed.connect(self.refresh_runs_button.click)
        # Respond to controls for the current run
        self.ui.export_button.clicked.connect(self.export_runs)
        # Set up 1D plotting widgets
        self.plot_1d_item = self.ui.plot_1d_view.getPlotItem()
        self.plot_2d_item = self.ui.plot_2d_view.getImageItem()
        self.plot_1d_item.addLegend()
        self.plot_1d_item.hover_coords_changed.connect(
            self.ui.hover_coords_label.setText
        )
        # Create a new export dialog for saving files
        self.export_dialog = ExportDialog(parent=self)

    def update_busy_hints(self):
        """Enable/disable UI elements based on the active hinters."""
        # Widgets for showing plots for runs
        if self._busy_hinters["run_widgets"] > 0:
            self.ui.detail_tabwidget.setEnabled(False)
        else:
            # Re-enable the run widgets
            self.ui.detail_tabwidget.setEnabled(True)
        # Widgets for selecting which runs to show
        if self._busy_hinters["run_table"] > 0:
            self.ui.run_tableview.setEnabled(False)
        else:
            # Re-enable the run widgets
            self.ui.run_tableview.setEnabled(True)
        # Widgets for filtering runs
        if self._busy_hinters["filters_widget"] > 0:
            self.ui.filters_widget.setEnabled(False)
        else:
            self.ui.filters_widget.setEnabled(True)
        # Update status message in message bars
        if len(list(self._busy_hinters.elements())) > 0:
            self.show_message("Loading…")
        else:
            self.show_message("Done.", 5000)

    @contextmanager
    def busy_hints(self, run_widgets=True, run_table=True, filter_widgets=True):
        """A context manager that displays UI hints when slow operations happen.

        Arguments can be used to control which widgets are modified.

        Usage:

        .. code-block:: python

            with self.busy_hints():
                self.db_task(self.slow_operation)

        Parameters
        ==========
        run_widgets
          Disable the widgets for viewing individual runs.
        run_table
          Disable the table for selecting runs to view.
        filter_widgets
          Disable the filter comboboxes, etc.

        """
        # Update the counters for keeping track of concurrent contexts
        hinters = {
            "run_widgets": run_widgets,
            "run_table": run_table,
            "filters_widget": filter_widgets,
        }
        hinters = [name for name, include in hinters.items() if include]
        self._busy_hinters.update(hinters)
        # Update the UI (e.g. disable widgets)
        self.update_busy_hints()
        # Run the innner context code
        try:
            yield
        finally:
            # Re-enable widgets if appropriate
            self._busy_hinters.subtract(hinters)
            self.update_busy_hints()

    @asyncSlot()
    @cancellable
    async def update_multi_signals(self, *args):
        """Retrieve a new list of signals for multi plot and update UI."""
        combobox = self.ui.multi_signal_x_combobox
        # Store old value for restoring later
        old_value = combobox.currentText()
        # Determine valid list of columns to choose from
        use_hints = self.ui.plot_multi_hints_checkbox.isChecked()
        signals_task = self.db_task(
            self.db.signal_names(hinted_only=use_hints), "multi signals"
        )
        xcols, ycols = await signals_task
        # Update the comboboxes with new signals
        combobox.clear()
        combobox.addItems(xcols)
        # Restore previous value
        combobox.setCurrentText(old_value)

    @asyncSlot()
    @cancellable
    async def update_1d_signals(self, *args):
        # Store old values for restoring later
        comboboxes = [
            self.ui.signal_x_combobox,
            self.ui.signal_y_combobox,
            self.ui.signal_r_combobox,
        ]
        old_values = [cb.currentText() for cb in comboboxes]
        # Determine valid list of columns to choose from
        use_hints = self.ui.plot_1d_hints_checkbox.isChecked()
        signals_task = self.db_task(
            self.db.signal_names(hinted_only=use_hints), "1D signals"
        )
        xcols, ycols = await signals_task
        self.multi_y_signals = ycols
        # Update the comboboxes with new signals
        self.ui.signal_x_combobox.clear()
        self.ui.signal_x_combobox.addItems(xcols)
        for cb in [
            self.ui.signal_y_combobox,
            self.ui.signal_r_combobox,
        ]:
            cb.clear()
            cb.addItems(ycols)
        # Restore previous values
        for val, cb in zip(old_values, comboboxes):
            cb.setCurrentText(val)

    @asyncSlot()
    @cancellable
    async def update_2d_signals(self, *args):
        # Store current selection for restoring later
        val_cb = self.ui.signal_value_combobox
        old_value = val_cb.currentText()
        # Determine valid list of dependent signals to choose from
        use_hints = self.ui.plot_2d_hints_checkbox.isChecked()
        xcols, vcols = await self.db_task(
            self.db.signal_names(hinted_only=use_hints), "2D signals"
        )
        # Update the UI with the list of controls
        val_cb.clear()
        val_cb.addItems(vcols)
        # Restore previous selection
        val_cb.setCurrentText(old_value)

    @asyncSlot()
    @cancellable
    async def update_multi_plot(self, *args):
        x_signal = self.ui.multi_signal_x_combobox.currentText()
        if x_signal == "":
            return
        use_hints = self.ui.plot_multi_hints_checkbox.isChecked()
        runs = await self.db_task(
            self.db.all_signals(hinted_only=use_hints), "multi-plot"
        )
        self.ui.plot_multi_view.plot_runs(runs, xsignal=x_signal)

    def update_export_button(self):
        # We can only export one scan at a time from here
        should_enable = self.selected_runs is not None and len(self.selected_runs) == 1
        self.ui.export_button.setEnabled(should_enable)

    @asyncSlot()
    async def export_runs(self):
        """Export the selected runs to user-specified filenames.

        Shows the user a file dialog with accepted types based on the
        accepted tiled export formats.

        """
        dialog = self.export_dialog
        # Determine default mimetypes
        mimetypes = self.selected_runs[0].formats()
        filenames = dialog.ask(mimetypes=mimetypes)
        mimetype = dialog.selectedMimeTypeFilter()
        formats = [mimetype] * len(filenames)
        await self.db_task(self.db.export_runs(filenames, formats=formats), "export")

    @asyncSlot()
    @cancellable
    async def update_1d_plot(self, *args):
        self.plot_1d_item.clear()
        # Figure out which signals to plot
        y_signal = self.ui.signal_y_combobox.currentText()
        x_signal = self.ui.signal_x_combobox.currentText()
        use_reference = self.ui.signal_r_checkbox.isChecked()
        if use_reference:
            r_signal = self.ui.signal_r_combobox.currentText()
        else:
            r_signal = None
        use_log = self.ui.logarithm_checkbox.isChecked()
        use_invert = self.ui.invert_checkbox.isChecked()
        use_grad = self.ui.gradient_checkbox.isChecked()
        task = self.db_task(
            self.db.signals(
                x_signal,
                y_signal,
                r_signal,
                use_log=use_log,
                use_invert=use_invert,
                use_grad=use_grad,
            ),
            "1D plot",
        )
        runs = await task
        self.ui.plot_1d_view.plot_runs(runs)

    @asyncSlot()
    @cancellable
    async def update_2d_plot(self):
        """Change the 2D map plot based on desired signals, etc."""
        # Figure out which signals to plot
        value_signal = self.ui.signal_value_combobox.currentText()
        use_log = self.ui.logarithm_checkbox_2d.isChecked()
        use_invert = self.ui.invert_checkbox_2d.isChecked()
        use_grad = self.ui.gradient_checkbox_2d.isChecked()
        images = await self.db_task(self.db.images(value_signal), "2D plot")
        # Get axis labels
        # Eventually this will be replaced with robus choices for plotting multiple images
        metadata = await self.db_task(self.db.metadata(), "2D plot")
        metadata = list(metadata.values())[0]
        dimensions = metadata["start"]["hints"]["dimensions"]
        try:
            xlabel = dimensions[-1][0][0]
            ylabel = dimensions[-2][0][0]
        except IndexError:
            # Not a 2D scan
            return
        # Get spatial extent
        extents = metadata["start"]["extents"]
        self.ui.plot_2d_view.plot_runs(
            images, xlabel=xlabel, ylabel=ylabel, extents=extents
        )

    @asyncSlot()
    async def update_metadata(self, *args):
        """Render metadata for the runs into the metadata widget."""
        # Combine the metadata in a human-readable output
        text = ""
        all_md = await self.db_task(self.db.metadata(), "metadata")
        for uid, md in all_md.items():
            text += f"# {uid}"
            text += yaml.dump(md)
            text += f"\n\n{'=' * 20}\n\n"
        # Update the widget with the rendered metadata
        self.ui.metadata_textedit.document().setPlainText(text)

    @asyncSlot()
    @cancellable
    async def update_selected_runs(self, *args):
        """Get the current runs from the database and stash them."""
        # Get UID's from the selection
        col_idx = self._run_col_names.index("UID")
        indexes = self.ui.run_tableview.selectedIndexes()
        uids = [i.siblingAtColumn(col_idx).data() for i in indexes]
        # Get selected runs from the database
        with self.busy_hints(run_widgets=True, run_table=False, filter_widgets=False):
            task = self.db_task(
                self.db.load_selected_runs(uids), "update selected runs"
            )
            self.selected_runs = await task
            # Update the necessary UI elements
            await self.update_multi_signals()
            await self.update_1d_signals()
            await self.update_2d_signals()
            await self.update_metadata()
            await self.update_1d_plot()
            await self.update_2d_plot()
            await self.update_multi_plot()
            self.update_export_button()

    def filters(self, *args):
        new_filters = {
            "proposal": self.ui.filter_proposal_combobox.currentText(),
            "esaf": self.ui.filter_esaf_combobox.currentText(),
            "sample": self.ui.filter_sample_combobox.currentText(),
            "exit_status": self.ui.filter_exit_status_combobox.currentText(),
            "use_current_proposal": bool(
                self.ui.filter_current_proposal_checkbox.checkState()
            ),
            "use_current_esaf": bool(self.ui.filter_current_esaf_checkbox.checkState()),
            "plan": self.ui.filter_plan_combobox.currentText(),
            "full_text": self.ui.filter_full_text_lineedit.text(),
            "edge": self.ui.filter_edge_combobox.currentText(),
            "user": self.ui.filter_user_combobox.currentText(),
        }
        null_values = ["", False]
        new_filters = {k: v for k, v in new_filters.items() if v not in null_values}
        return new_filters

    def load_models(self):
        # Set up the model
        self.runs_model = QStandardItemModel()
        # Add the model to the UI element
        self.ui.run_tableview.setModel(self.runs_model)

    def ui_filename(self):
        return "run_browser.ui"


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



from unittest import mock

import pytest
from bluesky_queueserver_api import BPlan
from ophyd_async.core import set_mock_value
from qtpy import QtCore

from firefly.plans.line_scan import LineScanDisplay


@pytest.fixture()
async def display(qtbot, sim_registry, sync_motors, async_motors, dxp, ion_chamber):
    display = LineScanDisplay()
    qtbot.addWidget(display)
    await display.update_devices(sim_registry)
    display.ui.run_button.setEnabled(True)
    return display


@pytest.mark.asyncio
async def test_time_calculator(display, sim_registry, ion_chamber, qtbot, qapp):
    # set up motor num
    await display.update_regions(2)

    # set up num of repeat scans
    display.ui.spinBox_repeat_scan_num.setValue(6)

    # set up scan num of points
    display.ui.scan_pts_spin_box.setValue(1000)

    # set up detectors
    display.ui.detectors_list.selected_detectors = mock.MagicMock(
        return_value=["vortex_me4", ion_chamber.name]
    )

    # set up default timing for the detector
    detectors = display.ui.detectors_list.selected_detectors()
    detectors = {name: sim_registry[name] for name in detectors}
    set_mock_value(ion_chamber.default_time_signal, 0.6255)
    detectors["vortex_me4"].default_time_signal.set(0.5).wait(2)

    # Trigger an update of the time calculator
    await display.update_total_time()

    # Check whether time is calculated correctly for a single scan
    assert display.ui.label_hour_scan.text() == "0"
    assert display.ui.label_min_scan.text() == "10"
    assert display.ui.label_sec_scan.text() == "25.5"

    # Check whether time is calculated correctly including the repeated scan
    assert display.ui.label_hour_total.text() == "1"
    assert display.ui.label_min_total.text() == "2"
    assert display.ui.label_sec_total.text() == "33.0"


@pytest.mark.asyncio
async def test_line_scan_plan_queued(qtbot, display):
    # set up motor num
    await display.update_regions(2)

    # set up a test motor 1
    display.regions[0].motor_box.combo_box.setCurrentText("async_motor_1")
    display.regions[0].start_line_edit.setText("1")
    display.regions[0].stop_line_edit.setText("111")

    # set up a test motor 2
    display.regions[1].motor_box.combo_box.setCurrentText("sync_motor_2")
    display.regions[1].start_line_edit.setText("2")
    display.regions[1].stop_line_edit.setText("222")

    # set up scan num of points
    display.ui.scan_pts_spin_box.setValue(10)

    # time is calculated when the selection is changed
    display.ui.detectors_list.selected_detectors = mock.MagicMock(
        return_value=["vortex_me4", "I0"]
    )

    # set up meta data
    display.ui.lineEdit_sample.setText("sam")
    display.ui.lineEdit_purpose.setText("test")
    display.ui.textEdit_notes.setText("notes")

    expected_item = BPlan(
        "scan",
        ["vortex_me4", "I0"],
        "async_motor_1",
        1.0,
        111.0,
        "sync_motor_2",
        2.0,
        222.0,
        num=10,
        md={"sample": "sam", "purpose": "test", "notes": "notes"},
    )

    def check_item(item):
        from pprint import pprint

        pprint(item.to_dict())
        pprint(expected_item.to_dict())
        return item.to_dict() == expected_item.to_dict()

    # Click the run button and see if the plan is queued
    with qtbot.waitSignal(
        display.queue_item_submitted, timeout=1000, check_params_cb=check_item
    ):
        qtbot.mouseClick(display.ui.run_button, QtCore.Qt.LeftButton)



import logging

from bluesky_queueserver_api import BPlan
from qtpy import QtWidgets

from firefly.component_selector import ComponentSelector
from firefly.plans import regions_display  # import RegionBase, RegionsDisplay

log = logging.getLogger(__name__)


class RobotMotorRegion(regions_display.RegionBase):
    def setup_ui(self):
        self.layout = QtWidgets.QHBoxLayout()

        # ComponentSelector
        self.motor_box = ComponentSelector()
        self.layout.addWidget(self.motor_box)

        # Motor position for robot transferring samples
        self.start_line_edit = QtWidgets.QLineEdit()
        self.start_line_edit.setPlaceholderText("Position…")
        self.layout.addWidget(self.start_line_edit)


class RobotDisplay(regions_display.RegionsDisplay):
    """Manage sample transfer using a robot plan.

    .. code-block:: python

      robot_sample(robot, number/None, motor1, position1, motor2,
                   position2, motor3, position3, …)

    """

    Region = RobotMotorRegion
    default_num_regions = 0

    def sample_numbers(self):
        sample_names = [name for name, device in self.device.samples.walk_subdevices()]
        sample_numbers = [int(name.replace("sample", "")) for name in sample_names]
        sample_numbers.append(None)
        return sample_numbers

    def customize_ui(self):
        super().customize_ui()
        # set the list of values for the combo box
        self.ui.sample_combo_box.clear()
        for sam in self.sample_numbers():
            self.ui.sample_combo_box.addItem(str(sam))

    def queue_plan(self, *args, **kwargs):
        """Execute this plan on the queueserver."""
        # Get scan parameters from widgets
        num_motor = self.ui.num_motor_spin_box.value()

        # Get the sample number from the sample_spin_box
        sam_num_str = self.ui.sample_combo_box.currentText()
        # Convert sam_num_str to an integer if it's a string representation of a number
        sam_num = int(sam_num_str) if sam_num_str.isdigit() else None

        # get parameters from motor regions
        motor_lst, position_lst = [], []
        for region_i in self.regions:
            motor_lst.append(region_i.motor_box.current_component().name)
            position_lst.append(float(region_i.start_line_edit.text()))

        args = [
            values for motor_i in zip(motor_lst, position_lst) for values in motor_i
        ]

        # Build the queue item
        robot = self.macros()["DEVICE"]
        item = BPlan("robot_transfer_sample", robot, sam_num, *args)

        # Submit the item to the queueserver
        log.info("Add ``robot_transfer_sample()`` plan to queue.")
        self.queue_item_submitted.emit(item)

    def ui_filename(self):
        return "robot.ui"


# -----------------------------------------------------------------------------
# :author:    Yanna Chen
# :email:     yannachen@anl.gov
# :copyright: Copyright © 2024, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full licens is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import pytest
from bluesky import RunEngine


class RunEngineStub(RunEngine):
    def __repr__(self):
        return "<run_engine.RunEngineStub>"


@pytest.fixture()
def RE(event_loop):
    return RunEngineStub(call_returns_result=True)



import numpy as np
import pandas as pd
import pytest
from tiled import queries

from haven.catalog import CatalogScan, unsnake


@pytest.fixture()
def scan(tiled_client):
    uid = "7d1daf1d-60c7-4aa7-a668-d1cd97e5335f"
    return CatalogScan(tiled_client[uid])


@pytest.fixture()
def grid_scan(tiled_client):
    uid = "85573831-f4b4-4f64-b613-a6007bf03a8d"
    return CatalogScan(tiled_client[uid])


def test_unsnake():
    # Make a snaked array
    arr = np.arange(27).reshape((3, 3, 3))
    snaked = np.copy(arr)
    snaked[::2] = snaked[::2, ::-1]
    snaked[:, ::2] = snaked[:, ::2, ::-1]
    # Do the unsnaking
    unsnaked = unsnake(snaked, [False, True, True])
    # Check the result
    np.testing.assert_equal(arr, unsnaked)


@pytest.mark.asyncio
async def test_client_fixture(tiled_client):
    """Does the client fixture load without stalling the test runner?"""


@pytest.mark.asyncio
async def test_load_scan(catalog):
    """Check that scans can be loaded from the catalog."""
    uid = "7d1daf1d-60c7-4aa7-a668-d1cd97e5335f"
    scan = await catalog[uid]
    assert isinstance(scan, CatalogScan)


@pytest.mark.asyncio
async def test_dataframe(scan):
    """Check that the catalogscan can produce a pandas dataframe."""
    df = await scan.to_dataframe()
    assert isinstance(df, pd.DataFrame)


@pytest.mark.asyncio
async def test_load_nd_data(grid_scan):
    """Check that the catalog scan can convert e.g. grid_scan results."""
    arr = await grid_scan["It_net_counts"]
    assert arr.ndim == 2
    assert arr.shape == (5, 21)


@pytest.mark.asyncio
async def test_distinct(catalog, tiled_client):
    distinct = tiled_client.distinct("plan_name")
    assert await catalog.distinct("plan_name") == distinct


@pytest.mark.asyncio
async def test_search(catalog, tiled_client):
    """Make sure we can query to database properly."""
    query = queries.Regex("plan_name", "xafs_scan")
    expected = tiled_client.search(query)
    response = await catalog.search(query)
    assert len(expected) == await response.__len__()


@pytest.mark.asyncio
async def test_values(catalog, tiled_client):
    """Get the individual scans in the catalog."""
    expected = [uid for uid in tiled_client.keys()]
    response = [val.uid async for val in catalog.values()]
    assert expected == response


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import logging
import time
import warnings

from apstools.plans.alignment import lineup
from bluesky import plan_stubs as bps
from bluesky.callbacks import best_effort
from bluesky.preprocessors import subs_wrapper

from ..preprocessors import shutter_suspend_decorator

log = logging.getLogger(__name__)


__all__ = ["align_motor", "align_pitch2"]


# @shutter_suspend_decorator()
def align_pitch2(
    pitch2, distance=200, reverse=False, detector="I0", bec=None, feature="cen", md={}
):
    """Tune the monochromator 2nd crystal pitch motor.

    Find and set the position of maximum intensity in the ion chamber
    "I0". The scanning range is relative to the current motor
    position, and will go *distance* above and below it. For example,
    if the current position is 1000, ``distance=200`` will scan from
    800 to 1200.

    Parameters
    ==========
    pitch2
      The pitch2 motor to use
    bec
      A bluesky best effort callback for finding the peak position.
    distance
      Relative distance to scan in either direction.
    reverse
      Whether the scan goes low-to-high (False) or high-to-low (True).
    detector
      Which detector name to use.
    feature
      Which feature of the peak to use for alignment.
    md
      Extra metadata to pass into the run engine.

    """
    md_ = dict(plan_name="align_pitch2")
    md_.update(md)
    # Prepare and run the plan
    yield from align_motor(
        detector=detector,
        motor=pitch2,
        distance=distance,
        reverse=reverse,
        bec=bec,
        feature=feature,
        md=md_,
    )


@shutter_suspend_decorator()
def align_motor(
    detector, motor, distance=200, reverse=False, bec=None, feature="cen", md={}
):
    """Center the given motor using the beam intensity.

    Find and set the position of maximum intensity in the ion chamber
    *detector*. The scanning range is relative to the current motor
    position, and will go *distance* above and below it. For example,
    if the current position is 1000, ``distance=200`` will scan from
    800 to 1200.

    Parameters
    ==========
    detector
      The detector device to use for measuring signals.
    motor
      The motor to use for alignment.
    bec
      A bluesky best effort callback for finding the peak position.
    distance
      Relative distance to scan in either direction.
    reverse
      Whether the scan goes low-to-high (False) or high-to-low (True).
    feature
      Which feature of the peak to use for alignment.
    md
      Extra metadata to pass into the run engine.

    """
    msg = (
        "The ``align_motor`` plan is deprecated. "
        "Consider using ``bluesky.plans.tune_centroid`` instead"
    )
    warnings.warn(msg, DeprecationWarning)
    # Prepare metadata
    md_ = dict(plan_name="align_motor")
    md_.update(md)
    # Set up the best effort callback
    if bec is None:
        bec = best_effort.BestEffortCallback()
        bec.disable_table()
    # Determine plan parameters
    start, end = (distance, -distance) if reverse else (-distance, distance)
    # Resolve motors and detectors
    detectors = [detector]
    if hasattr(detector, "raw_counts"):
        detectors.insert(0, detector.raw_counts)
    plan = lineup(
        detectors, motor, start, end, npts=40, feature=feature, bec=bec, md=md_
    )
    plan = subs_wrapper(plan, bec)
    yield from plan
    # Wait for the callback to catch up
    t0 = time.time()
    timeout = 5
    det_name = detectors[0].name
    while time.time() - t0 < timeout:
        new_value = bec.peaks[feature].get(det_name)
        if new_value is not None:
            if feature in ["max", "min"]:
                # Max and min also include the y-value
                new_value, y_value = new_value
            break
    # Set the motor to the new value (from below accounting for hysteresis)
    if new_value is None:
        # Didn't find a peak position
        msg = (
            f"No peak position found for {det_name},"
            f"motor '{motor.name}' will not be set."
        )
        log.error(msg)
        warnings.warn(msg)
    else:
        yield from bps.mv(motor, new_value - abs(end - start) / 2)
        yield from bps.mv(motor, new_value)


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import datetime as dt
import time
from unittest.mock import AsyncMock

import pytest
from pytestqt.exceptions import TimeoutError
from qtpy.QtCore import QTimer
from qtpy.QtWidgets import QAction

from firefly import queue_client

qs_status = {
    "msg": "RE Manager v0.0.18",
    "items_in_queue": 0,
    "items_in_history": 0,
    "running_item_uid": None,
    "manager_state": "idle",
    "queue_stop_pending": False,
    "queue_autostart_enabled": False,
    "worker_environment_exists": False,
    "worker_environment_state": "closed",
    "worker_background_tasks": 0,
    "re_state": None,
    "pause_pending": False,
    "run_list_uid": "4f2d48cc-980d-4472-b62b-6686caeb3833",
    "plan_queue_uid": "2b99ccd8-f69b-4a44-82d0-947d32c5d0a2",
    "plan_history_uid": "9af8e898-0f00-4e7a-8d97-0964c8d43f47",
    "devices_existing_uid": "51d8b88d-7457-42c4-b67f-097b168be96d",
    "plans_existing_uid": "65f11f60-0049-46f5-9eb3-9f1589c4a6dd",
    "devices_allowed_uid": "a5ddff29-917c-462e-ba66-399777d2442a",
    "plans_allowed_uid": "d1e907cd-cb92-4d68-baab-fe195754827e",
    "plan_queue_mode": {"loop": False},
    "task_results_uid": "159e1820-32be-4e01-ab03-e3478d12d288",
    "lock_info_uid": "c7fe6f73-91fc-457d-8db0-dfcecb2f2aba",
    "lock": {"environment": False, "queue": False},
}


devices_allowed = {
    "devices_allowed": {
        "cpt": {
            "classname": "Signal",
            "is_flyable": False,
            "is_movable": True,
            "is_readable": True,
            "module": "ophyd.signal",
        },
        "sim_detector": {
            "classname": "SynGauss",
            "components": {
                "Imax": {
                    "classname": "Signal",
                    "is_flyable": False,
                    "is_movable": True,
                    "is_readable": True,
                    "module": "ophyd.signal",
                },
                "center": {
                    "classname": "Signal",
                    "is_flyable": False,
                    "is_movable": True,
                    "is_readable": True,
                    "module": "ophyd.signal",
                },
                "noise": {
                    "classname": "EnumSignal",
                    "is_flyable": False,
                    "is_movable": True,
                    "is_readable": True,
                    "module": "ophyd.sim",
                },
                "noise_multiplier": {
                    "classname": "Signal",
                    "is_flyable": False,
                    "is_movable": True,
                    "is_readable": True,
                    "module": "ophyd.signal",
                },
                "sigma": {
                    "classname": "Signal",
                    "is_flyable": False,
                    "is_movable": True,
                    "is_readable": True,
                    "module": "ophyd.signal",
                },
                "val": {
                    "classname": "SynSignal",
                    "is_flyable": False,
                    "is_movable": True,
                    "is_readable": True,
                    "module": "ophyd.sim",
                },
            },
            "is_flyable": False,
            "is_movable": False,
            "is_readable": True,
            "module": "ophyd.sim",
        },
        "sim_detector_Imax": {
            "classname": "Signal",
            "is_flyable": False,
            "is_movable": True,
            "is_readable": True,
            "module": "ophyd.signal",
        },
        "sim_detector_center": {
            "classname": "Signal",
            "is_flyable": False,
            "is_movable": True,
            "is_readable": True,
            "module": "ophyd.signal",
        },
        "sim_detector_noise": {
            "classname": "EnumSignal",
            "is_flyable": False,
            "is_movable": True,
            "is_readable": True,
            "module": "ophyd.sim",
        },
        "sim_detector_noise_multiplier": {
            "classname": "Signal",
            "is_flyable": False,
            "is_movable": True,
            "is_readable": True,
            "module": "ophyd.signal",
        },
        "sim_detector_sigma": {
            "classname": "Signal",
            "is_flyable": False,
            "is_movable": True,
            "is_readable": True,
            "module": "ophyd.signal",
        },
        "sim_motor": {
            "classname": "SynAxis",
            "components": {
                "acceleration": {
                    "classname": "Signal",
                    "is_flyable": False,
                    "is_movable": True,
                    "is_readable": True,
                    "module": "ophyd.signal",
                },
                "readback": {
                    "classname": "_ReadbackSignal",
                    "is_flyable": False,
                    "is_movable": True,
                    "is_readable": True,
                    "module": "ophyd.sim",
                },
                "setpoint": {
                    "classname": "_SetpointSignal",
                    "is_flyable": False,
                    "is_movable": True,
                    "is_readable": True,
                    "module": "ophyd.sim",
                },
                "unused": {
                    "classname": "Signal",
                    "is_flyable": False,
                    "is_movable": True,
                    "is_readable": True,
                    "module": "ophyd.signal",
                },
                "velocity": {
                    "classname": "Signal",
                    "is_flyable": False,
                    "is_movable": True,
                    "is_readable": True,
                    "module": "ophyd.signal",
                },
            },
            "is_flyable": False,
            "is_movable": True,
            "is_readable": True,
            "module": "ophyd.sim",
        },
        "sim_motor_acceleration": {
            "classname": "Signal",
            "is_flyable": False,
            "is_movable": True,
            "is_readable": True,
            "module": "ophyd.signal",
        },
        "sim_motor_setpoint": {
            "classname": "_SetpointSignal",
            "is_flyable": False,
            "is_movable": True,
            "is_readable": True,
            "module": "ophyd.sim",
        },
        "sim_motor_unused": {
            "classname": "Signal",
            "is_flyable": False,
            "is_movable": True,
            "is_readable": True,
            "module": "ophyd.signal",
        },
        "sim_motor_velocity": {
            "classname": "Signal",
            "is_flyable": False,
            "is_movable": True,
            "is_readable": True,
            "module": "ophyd.signal",
        },
    },
    "devices_allowed_uid": "3664551b-368c-4a47-906a-b9f1ff6c8a91",
    "msg": "",
    "success": True,
}


@pytest.fixture()
def client():
    # Create a fake API with known responses
    api = AsyncMock()
    api.queue_start.return_value = {"success": True}
    api.status.return_value = qs_status
    api.queue_start.return_value = {
        "success": True,
    }
    api.re_resume.return_value = {
        "success": True,
    }
    api.re_stop.return_value = {
        "success": True,
    }
    api.re_abort.return_value = {
        "success": True,
    }
    api.re_halt.return_value = {
        "success": True,
    }
    api.devices_allowed.return_value = {"success": True, "devices_allowed": {}}
    api.environment_open.return_value = {"success": True}
    api.environment_close.return_value = {"success": True}
    api.queue_autostart.return_value = {"success": True}
    api.queue_stop.return_value = {"success": True}
    api.queue_stop_cancel.return_value = {"success": True}
    # Create the client using the fake API
    autoplay_action = QAction()
    autoplay_action.setCheckable(True)
    open_environment_action = QAction()
    open_environment_action.setCheckable(True)
    client = queue_client.QueueClient(api=api)
    yield client


def test_client_timer(client):
    assert isinstance(client.timer, QTimer)


@pytest.mark.asyncio
async def test_queue_re_control(client):
    """Test if the run engine can be controlled from the queue client."""
    api = client.api
    # Try and pause the run engine
    await client.request_pause(defer=True)
    # Check if the API paused
    api.re_pause.assert_called_once_with(option="deferred")
    # Pause the run engine now!
    api.reset_mock()
    await client.request_pause(defer=False)
    # Check if the API paused now
    api.re_pause.assert_called_once_with(option="immediate")
    # Start the queue
    api.reset_mock()
    await client.start_queue()
    # Check if the queue started
    api.queue_start.assert_called_once()
    # Resume a paused queue
    api.reset_mock()
    await client.resume_runengine()
    api.re_resume.assert_called_once()
    # Stop a paused queue
    api.reset_mock()
    await client.stop_runengine()
    api.re_stop.assert_called_once()
    # Abort a paused queue
    api.reset_mock()
    await client.abort_runengine()
    api.re_abort.assert_called_once()
    # Halt a paused queue
    api.reset_mock()
    await client.halt_runengine()
    api.re_halt.assert_called_once()


@pytest.mark.asyncio
async def test_run_plan(client, qtbot):
    """Test if a plan can be queued in the queueserver."""
    api = client.api
    api.item_add.return_value = {"success": True, "qsize": 2}
    new_status = qs_status.copy()
    new_status["items_in_queue"] = 2
    api.status.return_value = new_status
    # Send a plan
    with qtbot.waitSignal(
        client.length_changed, timeout=1000, check_params_cb=lambda l: l == 2
    ):
        await client.add_queue_item({})
    # Check if the API sent it
    api.item_add.assert_called_once_with(item={})


@pytest.mark.asyncio
async def test_toggle_autostart(client, qtbot):
    """Test how queuing a plan starts the runengine."""
    api = client.api
    # Check that it doesn't start the queue if the autoplay action is off
    assert not api.queue_autostart.called
    # Check the queue was started now that autoplay is on
    await client.toggle_autostart(True)
    api.queue_autostart.assert_called_once_with(True)


# def test_start_queue(ffapp, client, qtbot):
#     ffapp.start_queue_action.trigger()
#     qtbot.wait(1000)
#     client.api.queue_start.assert_called_once()


@pytest.mark.asyncio
async def test_stop_queue(client, qtbot):
    """Test how queuing a plan starts the runengine."""
    api = client.api
    # Check that it doesn't start the queue if the autoplay action is off
    assert not api.queue_autostart.called
    # Check the queue stop was requested
    await client.stop_queue(True)
    api.queue_stop.assert_called_once()
    # Check the queue stop can be cancelled
    api.clear_mock()
    await client.stop_queue(False)
    api.queue_stop_cancel.assert_called_once()


@pytest.mark.asyncio
async def test_check_queue_status(client, qtbot):
    # Check that the queue length is changed
    signals = [
        client.status_changed,
        client.environment_opened,
        client.environment_state_changed,
        client.re_state_changed,
        client.autostart_changed,
        client.manager_state_changed,
        client.in_use_changed,
    ]
    with qtbot.waitSignals(signals):
        await client.check_queue_status()
    return
    # Check that it isn't emitted a second time
    with pytest.raises(TimeoutError):
        with qtbot.waitSignals(signals, timeout=10):
            client.check_queue_status()
    # Now check a non-empty length queue
    new_status = qs_status.copy()
    new_status.update(
        {
            "worker_environment_exists": True,
            "worker_environment_state": "initializing",
            "manager_state": "creating_environment",
            "re_state": "idle",
            # "success": True,
            # "msg": "",
            # "items": ["hello", "world"],
            # "running_item": {},
            # "plan_queue_uid": "f682e6fa-983c-4bd8-b643-b3baec2ec764",
        }
    )
    client.api.status.return_value = new_status
    with qtbot.waitSignals(signals):
        client.check_queue_status()


@pytest.mark.asyncio
async def test_open_environment(client, qtbot):
    """Check that the 'open environment' action sends the right command to
    the queue.

    """
    api = client.api
    # Open the environment
    with qtbot.waitSignal(client.environment_opened) as blocker:
        await client.open_environment(True)
    assert blocker.args == [True]
    assert api.environment_open.called
    # Close the environment
    with qtbot.waitSignal(client.environment_opened) as blocker:
        await client.open_environment(False)
    assert blocker.args == [False]
    assert api.environment_close.called


@pytest.mark.asyncio
async def test_devices_available(client, qtbot):
    """Check that the queue client provides a list of devices that can be
    used in plans.

    """
    api = client.api
    api.devices_allowed.return_value = devices_allowed
    # Ask for updated list of devices
    with qtbot.waitSignal(client.devices_changed) as blocker:
        await client.update_devices()
    # Check that the data have the right form
    devices = blocker.args[0]
    assert "sim_detector" in devices.keys()


@pytest.mark.asyncio
async def test_update_status(client, time_machine, monkeypatch):
    api = client.api
    monkeypatch.setattr(
        queue_client, "load_config", lambda: {"beamline": {"hardware_is_present": True}}
    )
    # Set the last update timestamp to be long enough ago
    time_machine.move_to(dt.datetime(2024, 5, 29, 17, 51))
    client.last_update = time.monotonic()
    time_machine.move_to(dt.datetime(2024, 5, 29, 17, 55))
    client.timeout = 1
    # Update the client
    await client.update()
    assert api.status.called


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import logging
from typing import Mapping

from qtpy.QtCore import Signal
from qtpy.QtGui import QIcon, QKeySequence
from qtpy.QtWidgets import QAction, QMainWindow

log = logging.getLogger(__name__)


class Action(QAction):
    """An action with a few useful setup shortcuts."""

    window: QMainWindow = None

    def __init__(
        self,
        name: str,
        text: str,
        shortcut: str = None,
        icon: QIcon = None,
        tooltip: str = "",
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        # Set action properties
        self.setObjectName(name)
        self.setToolTip(tooltip)
        self.setText(text)
        if shortcut is not None:
            self.setShortcut(QKeySequence(shortcut))
        if icon is not None:
            self.setIcon(icon)

    @property
    def display(self):
        if self.window is not None:
            return self.window.display_widget()


class WindowAction(Action):
    """An action that opens a desired window when triggered."""

    window_created = Signal(QAction)
    window_shown = Signal(QAction)

    def __init__(
        self,
        name: str,
        text: str,
        display_file,
        shortcut: str = None,
        icon: QIcon = None,
        WindowClass: type = QMainWindow,
        macros=None,
        *args,
        **kwargs,
    ):
        super().__init__(
            name=name, text=text, shortcut=shortcut, icon=icon, *args, **kwargs
        )
        self.macros = macros
        self.display_file = display_file
        self.WindowClass = WindowClass
        # Connect signals
        self.triggered.connect(self.show_window)

    def show_window(self):
        if self.window is None:
            self.create_window()
        self.window.show()
        self.window.activateWindow()
        self.window_shown.emit(self)

    def forget_window(self):
        self.window = None

    def create_window(self):
        # Create the window
        window = self.WindowClass()
        self.window = window
        # Make it look pretty
        # window.update_tools_menu()
        kwargs = {}
        if self.macros is not None:
            kwargs["macros"] = self.macros
        window.open(str(self.display_file.resolve()), **kwargs)
        self.window_created.emit(self)
        # Properly remove the window if it's closed
        window.destroyed.connect(self.forget_window)


class ActionsRegistry:
    """A common namespace for keeping track of global actions."""

    # Actions for showing specific windows
    bss: WindowAction = None
    camera_overview: WindowAction = None
    energy: WindowAction = None
    iocs: WindowAction = None
    log: WindowAction = None
    status: WindowAction = None
    run_browser: WindowAction = None
    voltmeter: WindowAction = None
    xray_filter: WindowAction = None

    # Show windows for launching plans
    plans: Mapping[str, WindowAction]

    # Interactions with the queueserver
    queue_monitor: WindowAction = None
    queue_settings: Mapping[str, QAction]
    queue_controls: Mapping[str, QAction]

    # Show windows for controlling devices
    area_detectors: Mapping[str, WindowAction]
    cameras: Mapping[str, WindowAction]
    ion_chambers: Mapping[str, WindowAction]
    kb_mirrors: Mapping[str, WindowAction]
    mirrors: Mapping[str, WindowAction]
    motors: Mapping[str, WindowAction]
    robots: Mapping[str, WindowAction]
    slits: Mapping[str, WindowAction]
    tables: Mapping[str, WindowAction]
    xrf_detectors: Mapping[str, WindowAction]

    def __init__(self):
        self.plans = {}
        self.queue_settings = {}
        self.queue_controls = {}
        self.motors = {}
        self.cameras = {}
        self.area_detectors = {}
        self.slits = {}
        self.mirrors = {}
        self.ion_chambers = {}
        self.kb_mirrors = {}
        self.robots = {}
        self.tables = {}
        self.xrf_detectors = {}

    @property
    def all_actions(self):
        return [
            self.bss,
            self.camera_overview,
            self.energy,
            self.iocs,
            self.log,
            self.status,
            self.run_browser,
            self.voltmeter,
            self.xray_filter,
            *self.plans.values(),
            self.queue_monitor,
            *self.queue_settings.values(),
            *self.queue_controls.values(),
            *self.area_detectors.values(),
            *self.cameras.values(),
            *self.ion_chambers.values(),
            *self.kb_mirrors.values(),
            *self.mirrors.values(),
            *self.motors.values(),
            *self.robots.values(),
            *self.slits.values(),
            *self.tables.values(),
            *self.xrf_detectors.values(),
        ]



from unittest.mock import AsyncMock

import numpy as np
import pytest
from numpy.testing import assert_allclose
from ophyd_async.core import TriggerInfo, assert_value, get_mock_put, set_mock_value

from haven.devices.ion_chamber import IonChamber


@pytest.fixture()
def preamp(sim_ion_chamber):
    preamp = sim_ion_chamber.preamp
    return preamp


@pytest.fixture()
def ion_chamber(sim_registry):
    ion_chamber = IonChamber(
        scaler_prefix="255idcVME:3820:",
        scaler_channel=2,
        preamp_prefix="255idc:SR03:",
        voltmeter_prefix="255idc:LabjackT7_1:",
        voltmeter_channel=1,
        counts_per_volt_second=1e6,
        name="I0",
    )
    return ion_chamber


def test_ion_chamber_devices(ion_chamber):
    """Check that the ion chamber has the right sub-devices."""
    assert list(ion_chamber.mcs.scaler.channels.keys()) == [0, 2]
    assert hasattr(ion_chamber, "preamp")
    assert list(ion_chamber.voltmeter.analog_inputs.keys()) == [1]


async def test_readables(ion_chamber):
    await ion_chamber.connect(mock=True)
    expected_readables = [
        "I0-net_current",
        "I0-voltmeter-analog_inputs-1-final_value",
        "I0-mcs-scaler-channels-0-net_count",
        "I0-mcs-scaler-channels-0-raw_count",
        "I0-mcs-scaler-channels-2-net_count",
        "I0-mcs-scaler-channels-2-raw_count",
        "I0-mcs-scaler-elapsed_time",
    ]
    actual_readables = (await ion_chamber.describe()).keys()
    assert sorted(actual_readables) == sorted(expected_readables)
    # Check confirables
    expected_configables = [
        "I0-counts_per_volt_second",
        "I0-voltmeter-model_name",
        "I0-voltmeter-poll_sleep_ms",
        "I0-voltmeter-analog_in_sampling_rate",
        "I0-voltmeter-last_error_message",
        "I0-voltmeter-analog_in_resolution_all",
        "I0-voltmeter-firmware_version",
        "I0-voltmeter-device_temperature",
        "I0-voltmeter-serial_number",
        "I0-voltmeter-analog_in_settling_time_all",
        "I0-voltmeter-driver_version",
        "I0-voltmeter-ljm_version",
        "I0-voltmeter-analog_inputs-1-mode",
        "I0-voltmeter-analog_inputs-1-temperature_units",
        "I0-voltmeter-analog_inputs-1-low",
        "I0-voltmeter-analog_inputs-1-high",
        "I0-voltmeter-analog_inputs-1-resolution",
        "I0-voltmeter-analog_inputs-1-range",
        "I0-voltmeter-analog_inputs-1-differential",
        "I0-voltmeter-analog_inputs-1-enable",
        "I0-voltmeter-analog_inputs-1-input_link",
        "I0-voltmeter-analog_inputs-1-description",
        "I0-voltmeter-analog_inputs-1-scanning_rate",
        "I0-voltmeter-analog_inputs-1-device_type",
        "I0-mcs-model",
        "I0-mcs-lne_output_delay",
        "I0-mcs-count_on_start",
        "I0-mcs-lne_output_polarity",
        "I0-mcs-lne_output_stretcher",
        "I0-mcs-preset_time",
        "I0-mcs-channel_1_source",
        "I0-mcs-firmware",
        "I0-mcs-input_mode",
        "I0-mcs-output_mode",
        "I0-mcs-channel_advance_source",
        "I0-mcs-snl_connected",
        "I0-mcs-prescale",
        "I0-mcs-output_polarity",
        "I0-mcs-num_channels",
        "I0-mcs-input_polarity",
        "I0-mcs-lne_output_width",
        "I0-mcs-mux_output",
        "I0-mcs-acquire_mode",
        "I0-mcs-num_channels_max",
        "I0-mcs-dwell_time",
        "I0-mcs-scaler-channels-0-preset_count",
        "I0-mcs-scaler-channels-0-description",
        "I0-mcs-scaler-channels-0-is_gate",
        "I0-mcs-scaler-channels-0-offset_rate",
        "I0-mcs-scaler-channels-2-preset_count",
        "I0-mcs-scaler-channels-2-description",
        "I0-mcs-scaler-channels-2-is_gate",
        "I0-mcs-scaler-channels-2-offset_rate",
        "I0-mcs-scaler-delay",
        "I0-mcs-scaler-clock_frequency",
        "I0-mcs-scaler-preset_time",
        "I0-mcs-scaler-count_mode",
    ]
    actual_configables = (await ion_chamber.describe_configuration()).keys()
    assert sorted(actual_configables) == sorted(expected_configables)


@pytest.mark.asyncio
async def test_trigger(ion_chamber):
    await ion_chamber.connect(mock=True)
    assert ion_chamber._trigger_statuses == {}
    # Does the same trigger twice return the same
    status1 = ion_chamber.trigger()
    status2 = ion_chamber.trigger()
    assert not status1.done
    assert not status2.done
    await status1
    assert status1.done
    await status2
    assert status2.done


@pytest.mark.asyncio
async def test_trigger_dark_current(ion_chamber, monkeypatch):
    await ion_chamber.connect(mock=True)
    monkeypatch.setattr(
        ion_chamber.mcs.scaler.record_dark_current, "trigger", AsyncMock()
    )
    status = ion_chamber.trigger(record_dark_current=True)
    await status
    assert ion_chamber.mcs.scaler.record_dark_current.trigger.called


@pytest.mark.asyncio
async def test_net_current_signal(ion_chamber):
    """Test that scaler tick counts get properly converted to ion chamber current."""
    await ion_chamber.connect(mock=True)
    # Set the necessary dependent signals
    set_mock_value(ion_chamber.counts_per_volt_second, 10e6)  # 100 Mhz / 10 V
    set_mock_value(ion_chamber.scaler_channel.net_count, int(13e6))  # 1.3V
    set_mock_value(ion_chamber.mcs.scaler.clock_frequency, 9.6e6)
    set_mock_value(ion_chamber.mcs.scaler.channels[0].raw_count, 4.8e6)
    set_mock_value(ion_chamber.preamp.sensitivity_value, "20")
    set_mock_value(ion_chamber.preamp.sensitivity_unit, "uA/V")
    # Make sure it ignores the offset if it's off
    set_mock_value(ion_chamber.preamp.offset_on, "OFF")
    set_mock_value(ion_chamber.preamp.offset_value, "2")
    set_mock_value(ion_chamber.preamp.offset_unit, "uA")
    # Check the current answer
    assert (await ion_chamber.net_current.get_value()) == pytest.approx(5.2e-5)


@pytest.mark.asyncio
async def test_raw_current_signal(ion_chamber):
    """Test that scaler tick counts get properly converted to ion chamber current."""
    await ion_chamber.connect(mock=True)
    # Set the necessary dependent signals
    set_mock_value(ion_chamber.counts_per_volt_second, 10e6)  # 100 Mhz / 10 V
    set_mock_value(ion_chamber.scaler_channel.raw_count, int(13e6))  # 1.3V
    set_mock_value(ion_chamber.mcs.scaler.clock_frequency, 9.6e6)
    set_mock_value(ion_chamber.mcs.scaler.channels[0].raw_count, 4.8e6)
    set_mock_value(ion_chamber.preamp.sensitivity_value, "20")
    set_mock_value(ion_chamber.preamp.sensitivity_unit, "uA/V")
    # Make sure it ignores the offset if it's off
    set_mock_value(ion_chamber.preamp.offset_on, "OFF")
    set_mock_value(ion_chamber.preamp.offset_value, "2")
    set_mock_value(ion_chamber.preamp.offset_unit, "uA")
    # Check the current answer
    assert (await ion_chamber.raw_current.get_value()) == pytest.approx(5.2e-5)


async def test_voltmeter_name(ion_chamber):
    ion_chamber.auto_name = True
    await ion_chamber.connect(mock=True)
    assert (await ion_chamber.voltmeter_channel.description.get_value()) != "Icake"
    # Change the ion chamber name, and see if the voltmeter name updates
    set_mock_value(ion_chamber.scaler_channel.description, "Icake")
    await ion_chamber.connect(mock=True)
    assert (await ion_chamber.voltmeter_channel.description.get_value()) == "Icake"


def test_offset_pv(sim_registry):
    """Check that the device handles the weird offset numbering scheme.

    Net count PVs in the scaler go as

    - 25idcVME:3820:scaler1_netA.B
    - 25idcVME:3820:scaler1_netA.C
    - etc.

    but the offset PVs go
    - 25idcVME:3820:scaler1_offset0.B
    - ...
    - 25idcVME:3820:scaler1_offset0.D
    - 25idcVME:3820:scaler1_offset1.A
    - ...

    """
    channel_suffixes = [
        (1, "offset0.B"),
        (2, "offset0.C"),
        (3, "offset0.D"),
        (4, "offset1.A"),
        (5, "offset1.B"),
        (6, "offset1.C"),
        (7, "offset1.D"),
        (8, "offset2.A"),
        (9, "offset2.B"),
        (10, "offset2.C"),
        (11, "offset2.D"),
    ]
    for ch_num, suffix in channel_suffixes:
        ic = IonChamber(
            scaler_prefix="scaler_ioc:",
            scaler_channel=ch_num,
            preamp_prefix="",
            voltmeter_prefix="",
            voltmeter_channel=2,
            counts_per_volt_second=1,
            name=f"ion_chamber_{ch_num}",
        )
        assert (
            ic.scaler_channel.offset_rate.source == f"ca://scaler_ioc:scaler1_{suffix}"
        ), f"channel {ch_num}"


@pytest.fixture()
def trigger_info():
    return TriggerInfo(number=5, trigger="internal", deadtime=0, livetime=1.3)


async def test_flyscan_prepare(ion_chamber, trigger_info):
    # Prepare the ion chamber with mocked put commands
    await ion_chamber.connect(mock=True)
    set_mock_value(ion_chamber.mcs.num_channels_max, 8000)
    erase_mock_put = get_mock_put(ion_chamber.mcs.erase_all)
    assert not erase_mock_put.called
    # Prepare the ion chamber
    await ion_chamber.prepare(trigger_info)
    # Check that the device was properly configured for fly-scanning
    assert erase_mock_put.called
    await assert_value(ion_chamber.mcs.channel_advance_source, "Internal")
    await assert_value(ion_chamber.mcs.num_channels, 8000)
    await assert_value(ion_chamber.mcs.dwell_time, 1.3)
    assert ion_chamber._fly_readings == []


async def test_flyscan_kickoff(ion_chamber, trigger_info):
    await ion_chamber.connect(mock=True)
    await ion_chamber.prepare(trigger_info)
    # Prepare the mocked put commands
    start_mock_put = get_mock_put(ion_chamber.mcs.start_all)
    # Kickoff the fly scan
    status = ion_chamber.kickoff()
    set_mock_value(ion_chamber.mcs.acquiring, ion_chamber.mcs.Acquiring.ACQUIRING)
    await status
    # Check that the scan was started
    assert start_mock_put.called
    # Check that timestamps get recorded when new data are available
    set_mock_value(ion_chamber.mcs.current_channel, 0)
    assert len(ion_chamber._fly_readings) == 1


async def test_flyscan_complete(ion_chamber):
    await ion_chamber.connect(mock=True)
    # Run the complete method
    await ion_chamber.complete()
    # Check that the detector is stopped
    assert get_mock_put(ion_chamber.mcs.stop_all).called


async def test_flyscan_collect(ion_chamber, trigger_info):
    await ion_chamber.connect(mock=True)
    # Make fake fly-scan data
    sim_data = np.zeros(shape=(8000,))
    sim_data[:6] = [3, 5, 8, 13, 2, 33]
    sim_raw_data = np.zeros_like(sim_data)
    sim_raw_data[:6] = sim_data[:6] + 12
    set_mock_value(ion_chamber.mcs.mcas[2].spectrum, sim_raw_data)
    set_mock_value(ion_chamber.scaler_channel.offset_rate, 3)
    sim_times = np.asarray([4.0e7, 4.0e7, 4.0e7, 4.0e7, 4.0e7, 4.0e7])
    set_mock_value(ion_chamber.mcs.mcas[0].spectrum, sim_times)
    set_mock_value(ion_chamber.mcs.scaler.clock_frequency, 1e7)
    channel_numbers = range(len(sim_times) + 1)
    expected_timestamps = [1004, 1008, 1012, 1016, 1020, 1024]
    ion_chamber._fly_readings = [
        {
            ion_chamber.mcs.current_channel.name: {
                "value": datum,
                "timestamp": timestamp,
                "alarm_severity": 0,
            }
        }
        for (datum, timestamp) in zip(channel_numbers, expected_timestamps)
    ]
    # Ignore the first collected data point because it's during taxiing
    expected_data = sim_data[1:]
    # The real timestamps should be midway between PSO pulses
    collected = [c async for c in ion_chamber.collect_pages()]
    assert len(collected) == 1
    collected = collected[0]
    # Confirm data have the right structure
    raw_name = ion_chamber.scaler_channel.net_count.name
    assert collected["time"] == 1024
    assert_allclose(
        collected["data"][ion_chamber.scaler_channel.raw_count.name], sim_raw_data[:6]
    )
    assert_allclose(
        collected["data"][ion_chamber.scaler_channel.net_count.name], sim_data[:6]
    )
    assert_allclose(
        collected["data"][ion_chamber.mcs.scaler.elapsed_time.name], [4] * 6
    )
    assert_allclose(
        collected["timestamps"][ion_chamber.scaler_channel.raw_count.name],
        expected_timestamps,
    )
    assert_allclose(
        collected["timestamps"][ion_chamber.scaler_channel.net_count.name],
        expected_timestamps,
    )
    assert_allclose(
        collected["timestamps"][ion_chamber.mcs.scaler.elapsed_time.name],
        expected_timestamps,
    )


def test_default_time_signal(ion_chamber):
    assert ion_chamber.default_time_signal.source == "ca://255idcVME:3820:scaler1.TP"


@pytest.mark.asyncio
async def test_auto_naming_default(ion_chamber, monkeypatch):
    monkeypatch.setattr(
        ion_chamber.mcs.scaler.channels[2].description,
        "get_value",
        AsyncMock(return_value="I0"),
    )
    ion_chamber.auto_name = None
    ion_chamber.set_name("")
    await ion_chamber.connect(mock=True)
    assert ion_chamber.name == "I0"
    assert ion_chamber.mcs.name == "I0-mcs"


@pytest.mark.asyncio
async def test_auto_naming(ion_chamber, monkeypatch):
    monkeypatch.setattr(
        ion_chamber.mcs.scaler.channels[2].description,
        "get_value",
        AsyncMock(return_value="I0"),
    )
    ion_chamber.auto_name = True
    ion_chamber.set_name("")
    await ion_chamber.connect(mock=True)
    assert ion_chamber.name == "I0"
    assert ion_chamber.mcs.name == "I0-mcs"


@pytest.mark.asyncio
async def test_manual_naming(ion_chamber, monkeypatch):
    ion_chamber.set_name("")
    ion_chamber.auto_name = False
    await ion_chamber.connect(mock=True)
    assert ion_chamber.name == ""
    assert ion_chamber.mcs.name == ""


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



from unittest.mock import AsyncMock

import pytest
from bluesky.protocols import Flyable
from ophyd_async.core import get_mock_put

from haven.devices.motor import HavenMotor
from haven.devices.motor import Motor as AsyncMotor
from haven.devices.motor import load_motors


@pytest.fixture()
async def motor(sim_registry):
    motor = AsyncMotor("255idVME:m1", name="motor_1")
    await motor.connect(mock=True)
    return motor


@pytest.mark.asyncio
async def test_load_motors(sim_registry, monkeypatch):
    # Load the Ophyd motor definitions
    motors = load_motors(prefix="255idVME:", num_motors=3)
    # Were the motors imported correctly
    assert len(motors) == 3
    # assert type(motors[0]) is HavenMotor
    motor_names = [m.name for m in motors]
    assert "255idVME_m1" in motor_names
    assert "255idVME_m2" in motor_names
    assert "255idVME_m3" in motor_names
    # Check that the IOC name is set in labels
    motor1 = motors[0]
    assert "extra_motors" in motor1._ophyd_labels_


def test_motor_signals():
    m = HavenMotor("motor_ioc", name="test_motor")
    assert m.description.pvname == "motor_ioc.DESC"
    assert m.tweak_value.pvname == "motor_ioc.TWV"
    assert m.tweak_forward.pvname == "motor_ioc.TWF"
    assert m.tweak_reverse.pvname == "motor_ioc.TWR"
    assert m.soft_limit_violation.pvname == "motor_ioc.LVIO"


def test_async_motor_signals():
    m = AsyncMotor("motor_ioc", name="test_motor")
    assert m.description.source == "ca://motor_ioc.DESC"
    assert m.motor_is_moving.source == "ca://motor_ioc.MOVN"
    assert m.motor_done_move.source == "ca://motor_ioc.DMOV"
    assert m.high_limit_switch.source == "ca://motor_ioc.HLS"
    assert m.low_limit_switch.source == "ca://motor_ioc.LLS"
    assert m.high_limit_travel.source == "ca://motor_ioc.HLM"
    assert m.low_limit_travel.source == "ca://motor_ioc.LLM"
    assert m.direction_of_travel.source == "ca://motor_ioc.TDIR"
    assert m.soft_limit_violation.source == "ca://motor_ioc.LVIO"


def test_motor_flyer(motor):
    """Check that the haven motor implements the flyer interface."""
    assert motor is not None
    assert isinstance(motor, Flyable)


async def test_stop_button(motor):
    # Check that it got set up properly during __init__
    assert motor.motor_stop.name == "motor_1-motor_stop"
    assert motor.motor_stop.parent is motor
    await motor.motor_stop.trigger()
    mock = get_mock_put(motor.motor_stop)
    mock.assert_called_once_with(1, wait=False, timeout=10.0)


@pytest.mark.asyncio
async def test_auto_naming_default(monkeypatch):
    motor = AsyncMotor(prefix="255idVME:m1")
    monkeypatch.setattr(
        motor.description, "get_value", AsyncMock(return_value="motor_1")
    )
    await motor.connect(mock=True)
    assert motor.name == "motor_1"
    assert motor.user_setpoint.name == "motor_1-user_setpoint"


@pytest.mark.asyncio
async def test_auto_naming(monkeypatch):
    motor = AsyncMotor(prefix="255idVME:m1", name="not_the_final_name", auto_name=True)
    monkeypatch.setattr(
        motor.description, "get_value", AsyncMock(return_value="motor_1")
    )
    await motor.connect(mock=True)
    assert motor.name == "motor_1"
    assert motor.user_setpoint.name == "motor_1-user_setpoint"


@pytest.mark.asyncio
async def test_manual_naming(monkeypatch):
    motor = AsyncMotor(prefix="255idVME:m1", name="real_name", auto_name=False)
    await motor.connect(mock=True)
    assert motor.name == "real_name"
    assert motor.user_setpoint.name == "real_name-user_setpoint"


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import logging
import warnings
from pathlib import Path
from typing import Sequence

import qtawesome as qta
from pydm import data_plugins
from pydm.main_window import PyDMMainWindow
from qtpy import QtGui, QtWidgets

from haven import load_config

log = logging.getLogger(__name__)


class FireflyMainWindow(PyDMMainWindow):
    """A main window that will hold the various pydm displays.

    Parameters
    ==========
    actions
      A ActionsRegistry object. Will be used to run
      ``self.setup_menu_actions``.

    """

    hide_nav_bar: bool = False

    def __init__(
        self,
        actions=None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.setup_ui()
        if actions is not None:
            self.setup_menu_actions(actions)

    def setup_ui(self):
        # Hide the navbar initially
        self.ui.navbar.setVisible(False)
        # Add window icon
        root_dir = Path(__file__).parent.absolute()
        icon_path = root_dir / "splash.png"
        self.setWindowIcon(QtGui.QIcon(str(icon_path)))
        # Hide the nav bar
        if self.hide_nav_bar:
            self.toggle_nav_bar(False)
            self.ui.actionShow_Navigation_Bar.setChecked(False)
        # Prepare the status bar
        bar = self.statusBar()
        _label = QtWidgets.QLabel()
        _label.setText("Queue:")
        bar.addPermanentWidget(_label)
        self.ui.queue_length_label = QtWidgets.QLabel()
        self.ui.queue_length_label.setToolTip(
            "The length of the queue, not including the running plan."
        )
        self.ui.queue_length_label.setText("(??)")
        bar.addPermanentWidget(self.ui.queue_length_label)
        self.ui.environment_label = QtWidgets.QLabel()
        self.ui.environment_label.setToolTip(
            "The current state of the queue server environment."
        )
        self.ui.environment_label.setText("N/A")
        bar.addPermanentWidget(self.ui.environment_label)
        _label = QtWidgets.QLabel()
        _label.setText("/")
        bar.addPermanentWidget(_label)
        self.ui.re_label = QtWidgets.QLabel()
        self.ui.re_label.setToolTip("The current state of the queue server run engine.")
        self.ui.re_label.setText("N/A")
        bar.addPermanentWidget(self.ui.re_label)
        # Create menu bars (actual menu entries are set later)
        ## Setup menu
        self.ui.setup_menu = QtWidgets.QMenu(self.ui.menubar)
        self.ui.setup_menu.setObjectName("setup_menu")
        self.ui.setup_menu.setTitle("Set&up")
        self.ui.menubar.addAction(self.ui.setup_menu.menuAction())
        # Menu for managing the Queue server
        self.ui.queue_menu = QtWidgets.QMenu(self.ui.menubar)
        self.ui.queue_menu.setObjectName("menuQueue")
        self.ui.queue_menu.setTitle("&Queue")
        self.ui.menubar.addAction(self.ui.queue_menu.menuAction())
        ## Positioners menu
        self.ui.positioners_menu = QtWidgets.QMenu(self.ui.menubar)
        self.ui.positioners_menu.setObjectName("menuPositioners")
        self.ui.positioners_menu.setTitle("&Positioners")
        self.ui.menubar.addAction(self.ui.positioners_menu.menuAction())
        ## Scans menu
        self.ui.plans_menu = QtWidgets.QMenu(self.ui.menubar)
        self.ui.plans_menu.setObjectName("plans_menu")
        self.ui.plans_menu.setTitle("&Plans")
        self.ui.menubar.addAction(self.ui.plans_menu.menuAction())
        ## Detectors menu
        self.ui.detectors_menu = QtWidgets.QMenu(self.ui.menubar)
        self.ui.detectors_menu.setObjectName("detectors_menu")
        self.ui.detectors_menu.setTitle("&Detectors")
        self.ui.menubar.addAction(self.ui.detectors_menu.menuAction())
        # Connect signals to the status bar
        ...

    async def update_devices(self, registry):
        self.registry = registry
        await self.display_widget().update_devices(registry)

    def update_queue_controls(self, new_status):
        """Update the queue controls to match the state of the queueserver."""
        pass

    def update_queue_status(self, status):
        """Update the queue status labels."""
        self.ui.environment_label.setText(status["worker_environment_state"])
        new_length = status["items_in_queue"]
        self.ui.queue_length_label.setText(f"({new_length})")
        self.ui.re_label.setText(status["re_state"])
        # Notify the display of the new status
        display = self.display_widget()
        display.update_queue_status(status)

    def open(self, *args, **kwargs):
        widget = super().open(*args, **kwargs)
        # Connect signals for showing message in the window's status bar
        try:
            widget.status_message_changed.connect(self.show_status)
        except AttributeError:
            msg = (
                f"No status messages on window: {args}, {kwargs}. "
                "Possibly you're not using FireflyMainWindow "
                "or FireflyDisplay?"
            )
            log.warning(msg)
            warnings.warn(msg)
        # Add the caQtDM action to the menubar
        caqtdm_menu = self.ui.setup_menu
        caqtdm_actions = getattr(widget, "caqtdm_actions", [])
        if len(caqtdm_actions) > 0:
            caqtdm_menu.addSeparator()
        for action in caqtdm_actions:
            action.setIcon(qta.icon("fa5s.wrench"))
            caqtdm_menu.addAction(action)

    def closeEvent(self, event):
        super().closeEvent(event)
        # Delete the window so it's recreated next time it's opened
        self.deleteLater()

    def add_menu_action(self, action_name: str, text: str, menu: QtWidgets.QMenu):
        """Add a new QAction to a menubar menu.

        The action triggers when the menu item is activated.

        Parameters
        ==========
        action_name
          The name of the parameter to the save the action. Will be
          accessible via *action_name* on the window, and window.ui.
        text
          Human-readable text to show on the menu item.
        menu
          The QMenu object in which to put this menu item.

        Returns
        =======
        action
          A QAction for the menu item.

        """
        action = QtWidgets.QAction(self)
        action.setObjectName(action_name)
        action.setText(text)
        menu.addAction(action)
        # Save the action as object parameters
        setattr(self.ui, action_name, action)
        setattr(self, action_name, action)
        return action

    def setup_menu_actions(self, actions):
        self._setup_menu_actions(
            logs_window_action=actions.log,
            queue_monitor=actions.queue_monitor,
            queue_control_actions=actions.queue_controls,
            queue_settings_actions=actions.queue_settings,
            energy_window_action=actions.energy,
            filters_action=actions.xray_filter,
            slits_actions=actions.slits,
            mirror_actions=actions.mirrors,
            table_actions=actions.tables,
            robot_actions=actions.robots,
            plan_actions=actions.plans,
            run_browser_action=actions.run_browser,
            voltmeters_action=actions.voltmeter,
            motor_actions=actions.motors,
            ion_chamber_actions=actions.ion_chambers,
            camera_actions=actions.cameras,
            area_detector_actions=actions.area_detectors,
            xrf_detector_actions=actions.xrf_detectors,
            status_window_action=actions.status,
            bss_window_action=actions.bss,
            iocs_window_action=actions.iocs,
        )

    def _setup_menu_actions(
        self,
        logs_window_action,
        queue_monitor,
        queue_control_actions,
        queue_settings_actions,
        energy_window_action,
        filters_action,
        slits_actions,
        mirror_actions,
        table_actions,
        robot_actions,
        plan_actions,
        run_browser_action,
        voltmeters_action,
        motor_actions,
        ion_chamber_actions,
        camera_actions,
        area_detector_actions,
        xrf_detector_actions,
        status_window_action,
        bss_window_action,
        iocs_window_action,
    ):
        # Log viewer window
        if logs_window_action is not None:
            self.ui.menuView.addAction(logs_window_action)
        # Menu for managing the Queue server
        for action in queue_control_actions.values():
            self.ui.queue_menu.addAction(action)
        self.ui.queue_menu.addSeparator()
        # Queue settings for the queue client
        for action in queue_settings_actions.values():
            self.ui.queue_menu.addAction(action)
        self.ui.queue_menu.addAction(queue_monitor)
        # Sample viewer
        self.add_menu_action(
            action_name="actionShow_Sample_Viewer",
            text="Sample",
            menu=self.ui.positioners_menu,
        )
        # Motors sub-menu
        self.ui.motors_menu = QtWidgets.QMenu(self.ui.menubar)
        self.ui.motors_menu.setObjectName("motors_menu")
        self.ui.motors_menu.setTitle("Extra &Motors")
        motors_action = self.ui.motors_menu.menuAction()
        self.ui.positioners_menu.addAction(motors_action)
        motors_action.setIcon(qta.icon("mdi.cog-clockwise"))
        # Add actions to the motors sub-menus
        for action in motor_actions.values():
            self.ui.motors_menu.addAction(action)
        # Menu to launch the Window to change energy
        self.ui.positioners_menu.addAction(energy_window_action)
        # Add optical components
        if filters_action is not None:
            self.ui.positioners_menu.addAction(filters_action)
        if len(slits_actions) > 0:
            self.ui.positioners_menu.addSection("Slits")
        for action in slits_actions.values():
            self.ui.positioners_menu.addAction(action)
        if len(mirror_actions) > 0:
            self.ui.positioners_menu.addSection("Mirrors")
        for action in mirror_actions.values():
            self.ui.positioners_menu.addAction(action)
        if len(table_actions) > 0:
            self.ui.positioners_menu.addSection("Tables")
        for action in table_actions.values():
            self.ui.positioners_menu.addAction(action)
        if len(robot_actions) > 0:
            self.ui.positioners_menu.addSection("Robots")
        for action in robot_actions.values():
            self.ui.positioners_menu.addAction(action)
        # Add actions to the individual plans
        for action in plan_actions.values():
            self.ui.plans_menu.addAction(action)
        # Add entries for general scan management
        self.ui.plans_menu.addSeparator()
        if run_browser_action is not None:
            self.ui.plans_menu.addAction(run_browser_action)
        # Voltmeters window
        if voltmeters_action is not None:
            self.ui.detectors_menu.addAction(voltmeters_action)
        # Add an ion chamber sub-menu
        if len(ion_chamber_actions) > 0:
            self.ui.ion_chambers_menu = QtWidgets.QMenu(self.ui.menubar)
            self.ui.ion_chambers_menu.setObjectName("ion_chambers_menu")
            self.ui.ion_chambers_menu.setTitle("&Ion Chambers")
            self.ui.detectors_menu.addAction(self.ui.ion_chambers_menu.menuAction())
            # Add actions for the individual ion chambers
            for action in ion_chamber_actions.values():
                self.ui.ion_chambers_menu.addAction(action)
        # Cameras sub-menu
        self.ui.menuCameras = QtWidgets.QMenu(self.ui.menubar)
        self.ui.menuCameras.setObjectName("menuCameras")
        self.ui.menuCameras.setTitle("Cameras")
        self.ui.detectors_menu.addAction(self.ui.menuCameras.menuAction())
        # Add actions to the cameras sub-menus
        for action in camera_actions.values():
            self.ui.menuCameras.addAction(action)
        # Add area detectors to detectors menu
        if len(area_detector_actions) > 0:
            self.ui.detectors_menu.addSeparator()
        for action in area_detector_actions.values():
            self.ui.detectors_menu.addAction(action)
        # Add XRF detectors to detectors menu
        if len(xrf_detector_actions) > 0:
            self.ui.detectors_menu.addSeparator()
        for action in xrf_detector_actions.values():
            self.ui.detectors_menu.addAction(action)
        # Add other menu actions
        if status_window_action is not None:
            self.ui.menuView.addAction(status_window_action)
        if bss_window_action is not None:
            self.ui.setup_menu.addAction(bss_window_action)
        if iocs_window_action is not None:
            self.ui.setup_menu.addAction(iocs_window_action)
        # Make tooltips show up for menu actions
        for menu in [self.ui.setup_menu, self.ui.detectors_menu, self.ui.queue_menu]:
            menu.setToolTipsVisible(True)

    def show_status(self, message, timeout=0):
        """Show a message in the status bar."""
        bar = self.statusBar()
        bar.showMessage(message, timeout)

    def update_window_title(self):
        if self.showing_file_path_in_title_bar:
            title = self.current_file()
        else:
            title = self.display_widget().windowTitle()
        # Add the beamline name
        config = load_config()
        beamline_name = config["beamline"]["name"]
        title += f" - {beamline_name} - Firefly"
        if data_plugins.is_read_only():
            title += " [Read Only Mode]"
        self.setWindowTitle(title)


class PlanMainWindow(FireflyMainWindow):
    """A Qt window that has extra controls for a bluesky runengine."""

    navbar_actions: Sequence[str] = [
        "start",
        "|",
        "pause",
        "pause_now",
        "stop_queue",
        "|",
        "resume",
        "stop_runengine",
        "abort",
    ]

    hide_nav_bar: bool = True

    def setup_navbar(self, queue_control_actions):
        # Remove previous navbar actions
        navbar = self.ui.navbar
        for action in navbar.actions():
            navbar.removeAction(action)
        # Add runengine actions
        for key in self.navbar_actions:
            if key == "|":
                navbar.addSeparator()
            elif key in queue_control_actions:
                navbar.addAction(queue_control_actions[key])

    def update_queue_status(self, status):
        super().update_queue_status(status)
        # Apply style to the queue button (assuming it's called ``run_button``).
        display = self.display_widget()
        try:
            display.ui.run_button.update_queue_style(status)
        except AttributeError:
            pass

    def setup_menu_actions(self, actions):
        super().setup_menu_actions(actions=actions)
        self.setup_navbar(queue_control_actions=actions.queue_controls)

    def update_queue_controls(self, new_status):
        """Update the queue controls to match the state of the queueserver."""
        super().update_queue_controls(new_status)
        self.ui.navbar.setVisible(bool(new_status["in_use"]))


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------
